from __future__ import annotations

from typing import Optional

from PySide2.QtGui import QIcon
from PySide2.QtCore import Qt, Slot, Signal
from PySide2.QtWidgets import (QMenu, QLabel, QAction, QWidget, QLineEdit,
                               QHBoxLayout, QToolButton)

from ..widgets import ToolButton, show_help


class ClearTasksMenu(ToolButton):

    completed_tasks_removed = Signal()
    failed_tasks_removed = Signal()
    waiting_tasks_removed = Signal()
    all_tasks_removed = Signal()

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)

        self.setIcon(QIcon(':/icons/dark/trash'))
        self.setToolTip('Clear tasks options')
        self.setPopupMode(QToolButton.InstantPopup)
        self.setToolButtonStyle(Qt.ToolButtonIconOnly)

        self.clear_completed_tasks_btn = QAction('Clear Completed Tasks', self)
        self.clear_completed_tasks_btn.triggered.connect(self.completed_tasks_removed.emit)

        self.clear_failed_tasks_btn = QAction('Clear Failed Tasks', self)
        self.clear_failed_tasks_btn.triggered.connect(self.failed_tasks_removed.emit)

        self.clear_waiting_tasks_btn = QAction('Clear Waiting Tasks', self)
        self.clear_waiting_tasks_btn.triggered.connect(self.waiting_tasks_removed.emit)

        self.clear_all_tasks_btn = QAction('Clear All Tasks', self)
        self.clear_all_tasks_btn.triggered.connect(self.all_tasks_removed.emit)

        menu = QMenu(self)
        menu.addAction(self.clear_completed_tasks_btn)
        menu.addAction(self.clear_failed_tasks_btn)
        menu.addAction(self.clear_waiting_tasks_btn)
        menu.addAction(self.clear_all_tasks_btn)
        self.setMenu(menu)


class TableOptions(ToolButton):

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)

        self.setIcon(QIcon(':/icons/dark/settings-gear'))
        self.setToolTip('Table state')
        self.setPopupMode(QToolButton.InstantPopup)

        self.reset_layout = QAction('Reset Layout', self)
        self.resize_columns = QAction('Resize Columns', self)
        self.expand_all = QAction('Expand All', self)
        self.set_max_workers = QAction('Set Max Workers', self)

        self.help = QAction('Help', self)
        self.help.triggered.connect(show_help)

        menu = QMenu(self)
        menu.addAction(self.set_max_workers)
        menu.addSeparator()
        menu.addAction(self.reset_layout)
        menu.addAction(self.resize_columns)
        menu.addAction(self.expand_all)
        menu.addSeparator()
        menu.addAction(self.help)
        self.setMenu(menu)


class TasksViewToolbar(QWidget):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setObjectName('TqmTasksViewToolbar')

        self.clear_tasks = ClearTasksMenu()

        self.run_all_tasks = ToolButton()
        self.run_all_tasks.setIcon(QIcon(':/icons/dark/run-all'))
        self.run_all_tasks.setToolTip('Run all tasks')

        self.table_options = TableOptions()

        self.retry_all_btn = ToolButton()
        self.retry_all_btn.setIcon(QIcon(':/icons/dark/refresh'))
        self.retry_all_btn.setToolTip('Retry all failed tasks')

        self.search_btn = ToolButton()
        self.search_btn.setCheckable(True)
        self.search_btn.setIcon(QIcon(':/icons/dark/search'))
        self.search_btn.toggled.connect(self._on_toggle_search_bar)

        self.search_bar = QLineEdit()
        self.search_bar.setObjectName('TqmSearchBar')
        self.search_bar.setHidden(True)
        self.search_bar.setClearButtonEnabled(True)
        self.search_bar.setPlaceholderText('Search...')

        self.status_label = QLabel()
        self.status_label.setObjectName('TqmStatusLabel')

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.run_all_tasks)
        layout.addWidget(self.retry_all_btn)
        layout.addWidget(self.clear_tasks)
        layout.addWidget(self.table_options)
        layout.addWidget(self.search_btn)
        layout.addWidget(self.search_bar)
        layout.addStretch()
        layout.addWidget(self.status_label)

        self.setLayout(layout)

    @Slot(bool)
    def _on_toggle_search_bar(self, state: bool) -> None:
        focus = Qt.OtherFocusReason if state else Qt.NoFocusReason
        self.search_bar.setFocus(focus)
        self.search_bar.setVisible(state)



from __future__ import annotations

from typing import Union

from PySide2.QtGui import QBrush, QColor, QStandardItem, QLinearGradient
from PySide2.QtCore import Qt


class TaskItem(QStandardItem):
    def __init__(
        self,
        text: str,
        has_progress: bool = False,
        is_editable: bool = False,
        foreground: Union[QBrush, QColor, Qt.GlobalColor, QLinearGradient] = Qt.gray,
        alignment: Qt.AlignmentFlag = Qt.AlignLeft
    ):
        super().__init__(text)
        self.setTextAlignment(alignment)
        self.setEditable(is_editable)
        self.setForeground(foreground)

        if has_progress:
            # set initial progress to 0
            self.setData(0, Qt.DisplayRole)



from .open_file import open_file
from .thread_info import get_thread_info
from .random_color import RandomColor
from .extract_fn_name import extract_fn_name

__all__ = [
    'extract_fn_name',
    'RandomColor',
    'get_thread_info',
    'open_file'
]



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from PySide2.QtCore import Slot, Signal, QObject, QRunnable

from ..exceptions import TaskEventError
from .task_callbacks import TaskEventCallbacks

if TYPE_CHECKING:
    from .task import TaskUnit, TaskGroup, TaskExecutable


def execute_user_callback(callback: TaskEventCallbacks, obj: TaskUnit) -> None:
    if getattr(obj.callbacks, callback):
        getattr(obj.callbacks, callback)(obj)


class RunnerSignals(QObject):
    """
    Signals emitted by the BaseRunner class.
    """

    runner_completed = Signal(object)
    runner_failed = Signal(object, Exception)
    runner_started = Signal(object)
    runner_finished = Signal(object)

    # progress is used by the user if needed
    task_progress_updated = Signal(int)

    # when group needs to add its tasks
    group_task_added = Signal(object)


class BaseRunner(QRunnable):

    def __init__(self) -> None:
        super().__init__()
        self.signals = RunnerSignals()
        self.setAutoDelete(True)

    def __str__(self) -> str:
        return f'<{self.__class__.__name__} {hex(id(self))}>'


class GroupRunner(BaseRunner):
    """
    Represents a task that can be executed in a separate thread.
    """

    def __init__(self, group: TaskGroup) -> None:
        super().__init__()
        self.group = group

    @Slot()
    def run(self) -> None:

        execute_user_callback(TaskEventCallbacks.ON_START, obj=self.group)
        self.signals.runner_started.emit(self.group)

        # add all tasks from the main thread
        self.signals.group_task_added.emit(self.group.tasks)

        # Block until all tasks are complete
        while not all(task.state.is_completed or task.state.is_failed for task in self.group.tasks):
            time.sleep(0.1)  # sleep to avoid CPU thrashing

        if all(task.state.is_completed for task in self.group.tasks):
            self.signals.runner_completed.emit(self.group)
            execute_user_callback(TaskEventCallbacks.ON_COMPLETED, obj=self.group)

        elif any(task.state.is_failed for task in self.group.tasks):
            self.signals.runner_failed.emit(self.group, TaskEventError('Some tasks failed'))
            execute_user_callback(TaskEventCallbacks.ON_FAILED, obj=self.group)

        self.signals.runner_finished.emit(self.group)
        execute_user_callback(TaskEventCallbacks.ON_FINISH, obj=self.group)


class TaskRunner(BaseRunner):
    """
    Represents a task that can be executed in a separate thread.
    """

    def __init__(self, task: TaskExecutable) -> None:
        super().__init__()
        self.task = task

    @Slot()
    def run(self) -> None:

        execute_user_callback(TaskEventCallbacks.ON_START, obj=self.task)

        try:
            self.signals.runner_started.emit(self.task)
            self.task.execute(self.task)

        except Exception as e:
            self.signals.runner_failed.emit(self.task, e)
            execute_user_callback(TaskEventCallbacks.ON_FAILED, obj=self.task)

        else:
            self.signals.runner_completed.emit(self.task)
            execute_user_callback(TaskEventCallbacks.ON_COMPLETED, obj=self.task)

        finally:
            self.signals.runner_finished.emit(self.task)
            execute_user_callback(TaskEventCallbacks.ON_FINISH, obj=self.task)



from __future__ import annotations

from typing import Any, Dict, List, Tuple, Union, Optional
from functools import partial

from PySide2.QtGui import (QIcon, QStandardItem, QContextMenuEvent,
                           QStandardItemModel)
from PySide2.QtCore import Qt, Slot, Signal, QModelIndex, QSortFilterProxyModel
from PySide2.QtWidgets import (QMenu, QAction, QWidget, QSplitter, QTreeView,
                               QVBoxLayout)

from .toolbar import TasksViewToolbar
from ..widgets import Frame
from .task_item import TaskItem
from .._core.task import TaskUnit, TaskGroup, TaskExecutable
from ..exceptions import TaskGroupError
from .font_loader import get_monospace_font
from .context_menu import TaskFileMenu
from .debug_widget import DebugWidget
from .mixins.view_mixin import ViewStateMixing
from .._core.task_actions import TaskAction, TaskActionVisibility
from .._core.task_executor import TaskExecutor
from .progress_bar_delegate import ProgressBarDelegate
from .mixins.multi_select_mixin import MultiSelectMixin

_TaskActionsT = Union[
    Tuple[TaskAction[TaskExecutable], ...],
    Tuple[TaskAction[TaskGroup], ...]
]


class TreeModel(QStandardItemModel):

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self._headers = (
            'Task', 'Progress', 'Wait for', 'Comment',
            'Created', 'Started', 'Completed',
        )

        self.columns = {h: i for i, h in enumerate(self._headers)}

        self.setHorizontalHeaderLabels(self._headers)
        self.root_item = self.invisibleRootItem()

        self.itemChanged.connect(self._on_item_changed)

    @Slot(TaskItem)
    def _on_item_changed(self, item: TaskItem) -> None:
        if item.column() == self.columns['Comment']:
            task = self.itemFromIndex(item.index().siblingAtColumn(0)).data(Qt.UserRole)
            task.comment = item.text()

    def _create_task_row(self, item: TaskItem, task: TaskUnit) -> List[TaskItem]:
        return [
            item,
            TaskItem('', has_progress=True),
            TaskItem(task.parent.name if task.parent else ''),
            TaskItem(task.comment, is_editable=True),
            TaskItem(task.state.get_first().timestamp, alignment=Qt.AlignCenter),
            TaskItem('', alignment=Qt.AlignCenter),
            TaskItem('', alignment=Qt.AlignCenter),
        ]

    def _configure_group_item(self, task_group: TaskGroup) -> None:
        if not task_group.tasks:
            raise TaskGroupError(
                f'Cannot add TaskGroup "{task_group.name}" without any tasks'
            )

        task_group.progress_bar.maximum = len(task_group.tasks)

        group_item = task_group.item
        if not group_item:
            raise TaskGroupError(
                f'Expected TaskGroup "{task_group.name}" to have an associated group item, but none was found.'
            )

        group_item.setIcon(QIcon(':/icons/dark/circle-parent'))
        f = group_item.font()
        f.setBold(True)
        group_item.setFont(f)

        for task in task_group.tasks:
            task.color = task_group.color

            task.item = TaskItem(task.name, foreground=task.color)
            task.item.setIcon(QIcon(':/icons/dark/circle-line'))
            task.item.setData(task, Qt.UserRole)

            group_item.appendRow(self._create_task_row(task.item, task))

    def add_task(self, task_unit: TaskUnit) -> TaskItem:

        task_unit.item = TaskItem(task_unit.name, foreground=task_unit.color)
        task_unit.item.setIcon(QIcon(':/icons/dark/circle-filled'))
        task_unit.item.setData(task_unit, Qt.UserRole)

        if isinstance(task_unit, TaskGroup):
            self._configure_group_item(task_unit)

        self.root_item.appendRow(self._create_task_row(task_unit.item, task_unit))

        return task_unit.item

    def reset(self) -> None:
        self.clear()
        self.setHorizontalHeaderLabels(self._headers)
        self.root_item = self.invisibleRootItem()


class RecursiveFilterProxyModel(QSortFilterProxyModel):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:

        if super().filterAcceptsRow(source_row, source_parent):
            return True

        model = self.sourceModel()
        source_index = model.index(source_row, 0, source_parent)

        if source_index.isValid():
            rows = model.rowCount(source_index)
            for i in range(rows):
                if self.filterAcceptsRow(i, source_index):
                    return True
        return False


class TaskTreeView(MultiSelectMixin, ViewStateMixing, QTreeView):
    selected_task_retried = Signal(TaskExecutable)
    selected_task_removed = Signal()

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(self, self, parent)
        self.setObjectName('TqmTasksTreeView')

        self.tasks_model = TreeModel()

        self.proxy_model = RecursiveFilterProxyModel()
        self.proxy_model.setFilterKeyColumn(-1)
        self.proxy_model.setSortCaseSensitivity(Qt.CaseInsensitive)
        self.proxy_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.proxy_model.setSourceModel(self.tasks_model)
        self.setModel(self.proxy_model)

        self._progress_delegate = ProgressBarDelegate(self.tasks_model.columns['Progress'], self)
        self.setItemDelegate(self._progress_delegate)

        self.setAlternatingRowColors(True)
        self.setFont(get_monospace_font())

        self.setSelectionMode(QTreeView.SingleSelection)
        self.setSelectionBehavior(QTreeView.SelectRows)

        header = self.header()
        header.setObjectName('TqmHeaderView')
        header.setSortIndicatorShown(True)
        header.setStretchLastSection(True)
        header.setSectionsMovable(True)

        self.load_table_state()
        self.setSortingEnabled(True)

    def stop_animation(self) -> None:
        self._progress_delegate.cleanup()

    def get_column_indexes(self) -> Dict[str, int]:
        """Return a dictionary of column names and their corresponding indexes."""
        return self.tasks_model.columns

    def get_column_index(self, column_name: str) -> int:
        return self.tasks_model.columns[column_name]

    def get_selected_items(self) -> List[QStandardItem]:
        """Get a list of selected tasks QStandardItem in the tree view."""
        items: List[QStandardItem] = []
        for index in self.selectionModel().selectedRows():

            source_index = self.proxy_model.mapToSource(index)
            item = self.tasks_model.itemFromIndex(source_index)

            items.append(item)

        return items

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        def add_actions(actions: _TaskActionsT, visibility: TaskActionVisibility) -> None:
            for action in filter(lambda a: a.visibility == visibility, actions):
                if action.name == '%file%':
                    # get the file assigned to the action return
                    user_file = action.action(task)
                    menu.addMenu(TaskFileMenu(user_file, self))
                else:
                    act = QAction(action.name, self)
                    act.triggered.connect(partial(action.action, task))
                    menu.addAction(act)

        item = self.indexAt(event.pos()).siblingAtColumn(0)
        if not item.isValid() or not item.data(Qt.UserRole):
            return

        menu = QMenu(self)

        task: TaskUnit = item.data(Qt.UserRole)
        add_actions(task.actions, TaskActionVisibility.ALWAYS)

        if task.state.is_completed:
            add_actions(task.actions, TaskActionVisibility.ON_COMPLETED)

        elif task.state.is_failed:
            add_actions(task.actions, TaskActionVisibility.ON_FAILED)

            if isinstance(task, TaskExecutable) and not task.parent:
                retry_act = QAction('Retry', self)
                retry_act.triggered.connect(lambda: self.selected_task_retried.emit(task))
                menu.addAction(retry_act)

        menu.addSeparator()

        if task.state.is_removable:
            remove_task = QAction('Remove selected', self)
            remove_task.triggered.connect(self.selected_task_removed.emit)
            menu.addAction(remove_task)

        menu.exec_(event.globalPos())


class TaskManagerView(QWidget):
    """
    A widget that displays tasks and provides various methods to interact with them.

    This widget consists of a tree view to display tasks, a search bar, buttons for
    actions, a log widget to display logs, and a status bar to show the current status.

    """

    def __init__(self, executor: TaskExecutor, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.tree_view = TaskTreeView()
        self.tree_view.clicked.connect(self._on_toggle_debug)

        self.toolbar = TasksViewToolbar()
        self.toolbar.search_bar.textChanged.connect(
            self.tree_view.proxy_model.setFilterRegExp
        )

        self._debug_widget = DebugWidget(executor)

        table_layout = QVBoxLayout()
        table_layout.addWidget(self.toolbar)
        table_layout.addWidget(self.tree_view)

        table_widget = Frame()
        table_widget.setLayout(table_layout)

        splitter = QSplitter()
        splitter.addWidget(table_widget)
        splitter.addWidget(self._debug_widget)
        splitter.setSizes([1000, 0])

        layout = QVBoxLayout()
        layout.addWidget(splitter)
        self.setLayout(layout)

    @Slot(QModelIndex)
    def _on_toggle_debug(self, index: QModelIndex) -> None:
        selected_item = self.tree_view.get_selected_items()
        task = selected_item[0].data(Qt.UserRole)
        self._debug_widget.populate(task)

    def toggle_expand(self, state: bool) -> None:
        self.tree_view.expandAll() if state else self.tree_view.collapseAll()

    def update_status(self, status: Dict[str, Any]) -> None:
        stringified_status = ' | '.join(f'{k}: {v}' for k, v in status.items())
        self.toolbar.status_label.setText(stringified_status)



"""ThreadPoolInterface and ThreadPoolWrapper classes.

This module only exists to simplify testing by providing a
ThreadPoolInterface that can be mocked.
"""
from __future__ import annotations

from abc import ABC, abstractmethod

from PySide2.QtCore import QObject, QRunnable, QThreadPool


class ThreadPoolInterface(ABC):
    @abstractmethod
    def start(self, runnable: QRunnable, priority: int = ...) -> None: ...
    @abstractmethod
    def activeThreadCount(self) -> int: ...
    @abstractmethod
    def waitForDone(self) -> None: ...
    @abstractmethod
    def setMaxThreadCount(self, max_thread_count: int) -> None: ...


class ThreadPoolWrapper(ThreadPoolInterface):
    def __init__(self, parent: QObject):
        self._threadpool = QThreadPool(parent)

    def start(self, runnable: QRunnable, priority: int = 0) -> None:
        self._threadpool.start(runnable, priority)

    def activeThreadCount(self) -> int:
        return self._threadpool.activeThreadCount()

    def waitForDone(self) -> None:
        self._threadpool.waitForDone()

    def setMaxThreadCount(self, max_thread_count: int) -> None:
        self._threadpool.setMaxThreadCount(max_thread_count)




from __future__ import annotations

from typing import Any, cast

from PySide2.QtGui import QKeyEvent, QMouseEvent
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QAbstractItemView


class MultiSelectMixin:
    """
    Mixin for multi-selecting items in a QAbstractItemView.

    This mixin provides methods for enabling multi-selection of items in a QAbstractItemView.
    It allows selecting multiple items by holding the Shift key and selecting all items by
    pressing the 'A' key.

    Args:
        view (QAbstractItemView): The view to enable multi-selection for.

    """

    def __init__(self, view: QAbstractItemView, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self.view = view
        self.view.setSelectionMode(QAbstractItemView.SingleSelection)

    @property
    def parent(self) -> QAbstractItemView:
        return cast(QAbstractItemView, super())

    def mousePressEvent(self, event: QMouseEvent) -> None:
        if (
            event.button() == Qt.RightButton
            and self.view.selectionMode() == QAbstractItemView.SingleSelection
        ):
            return
        return self.parent.mousePressEvent(event)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """
        Handle key press events.

        This method handles key press events and enables multi-selection mode when the Shift key
        is pressed. It also allows selecting all items when the 'A' key is pressed.

        Args:
            event (QKeyEvent): The key press event.

        Returns:
            None

        """
        if event.modifiers() == Qt.ShiftModifier:
            self.view.setSelectionMode(QAbstractItemView.MultiSelection)

        elif event.key() == Qt.Key_A:
            if self.view.selectionMode() == QAbstractItemView.SingleSelection:
                self.view.setSelectionMode(QAbstractItemView.MultiSelection)
                self.view.selectAll()
            else:
                self.view.clearSelection()
                self.view.setSelectionMode(QAbstractItemView.SingleSelection)

        self.parent.keyPressEvent(event)

    def keyReleaseEvent(self, event: QKeyEvent) -> None:
        """
        Handle key release events.

        This method handles key release events and restores the selection mode to single selection
        when the Shift key is released.

        Args:
            event (QKeyEvent): The key release event.

        Returns:
            None

        """
        if event.key() == Qt.Key_Shift:
            self.view.setSelectionMode(QAbstractItemView.SingleSelection)
        self.parent.keyReleaseEvent(event)



from __future__ import annotations

import os
import shlex
import platform
import subprocess


def open_file(file_path: str, reveal: bool = False) -> None:
    """
    Opens the specified file using the default application associated with the file type.

    Args:
        file_path (str): The path of the file to be opened.
        reveal (bool, optional): If True, the file will be revealed in the file
        explorer. Defaults to False.

    Raises:
        Exception: If the operating system is not supported.

    """
    if not os.path.exists(file_path):
        return

    if platform.system() == 'Darwin':
        subprocess.run(
            shlex.split(f'open -R "{file_path}"' if reveal else f'open "{file_path}"')
        )

    elif platform.system() == 'Windows':
        if reveal:
            subprocess.run(f'explorer /select,"{file_path}"', shell=True)
        else:
            os.startfile(file_path)

    elif platform.system() == 'Linux':
        subprocess.run(
            ['xdg-open', os.path.dirname(file_path) if reveal else file_path]
        )

    else:
        raise OSError(f'{platform.system()} is not supported.')



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict
from functools import partial

from PySide2.QtCore import Qt, QPoint, QTimer, QByteArray
from PySide2.QtWidgets import QMenu, QAction

from ..._core.settings import open_settings

if TYPE_CHECKING:
    from ..._ui.ui_view_model import TaskTreeView


class ViewStateMixing:
    """
    Mixin for handling the persistent state of a view.

    This mixin offers functionality to save and restore the state of a view, including
    column widths and visibility. It also provides a context menu on the view's header
    to allow users to customize which columns are visible.

    Args:
        view (TaskTreeView): The view whose state will be managed.

    Note:
        This class contains extra complexity due to legacy support for both TaskTreeView
        and TaskDatabaseView. Consider refactoring if only one view type is needed in the future.
    """

    def __init__(
        self,
        view: TaskTreeView,
        *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        self.__view = view

        self.__header = self.__view.header()
        self.__parent = 'tasks'

        self.__initial_state = self.__header.saveState()
        self.__columns: Dict[str, int] = {}

        self.__timer = QTimer()
        self.__timer.setInterval(2000)
        self.__timer.timeout.connect(self._save_table_state)
        self.__timer.setSingleShot(True)

        self.__header.sectionMoved.connect(self._activate_timer)
        self.__header.sectionResized.connect(self._activate_timer)
        self.__header.setContextMenuPolicy(Qt.CustomContextMenu)
        self.__header.customContextMenuRequested.connect(self._on_header_menu)

        with open_settings(mode='w') as settings:
            if self.__parent not in settings.view:
                settings.view[self.__parent] = {
                    'state': '',
                    'columns': {}
                }

    def _activate_timer(self) -> None:
        if not self.__timer.isActive():
            self.__timer.start()

    def get_columns(self) -> Dict[str, int]:
        """
        Map the column names to their respective indices.

        Returns:
            dict[str, int]: A dictionary mapping column names to their indices.

        """
        if not self.__columns:
            self.__columns = self.__view.get_column_indexes()
        return self.__columns

    def load_table_state(self) -> None:
        """
        Loads the state of the table from the settings and applies it to the view.

        This method reads the table state from the settings, restores the header state,
        and hides or shows columns based on the saved state. If no table state is found
        in the settings, it uses the initial state.

        Returns:
            None

        """
        with open_settings() as s:
            settings = s

        current_state = settings.view[self.__parent]['state']

        state = QByteArray.fromBase64(settings.view[self.__parent]['state'].encode())

        if state.isEmpty():
            state = self.__initial_state

        self.__header.restoreState(state)

        for column, index in self.get_columns().items():
            self.__view.setColumnHidden(
                index, not settings.view[self.__parent]['columns'].get(column, True)
            )

        # when running first time, we want to set the size automatically
        if not current_state:
            for col in range(self.__view.tasks_model.columnCount()):
                self.__view.setColumnWidth(col, 250)

    def _save_table_state(self) -> None:
        """
        Save the state of the table to the settings.

        This method saves the current state of the table, including the column widths
        and visibility, to the settings.

        Returns:
            None

        """
        with open_settings(mode='w') as settings:
            settings.view[self.__parent]['state'] = (
                self.__header.saveState().toBase64().data().decode()
            )

    def _update_columns(self, column_name: str, state: bool) -> None:
        """
        Update the visibility of a column.

        This method updates the visibility of a column based on the provided state.

        Args:
            column_name (str): The name of the column to update.
            state (bool): The new visibility state of the column.

        Returns:
            None

        """
        with open_settings(mode='w') as settings:
            settings.view[self.__parent]['columns'][column_name] = state
            self.__view.setColumnHidden(self.get_columns()[column_name], not state)

    def _on_header_menu(self, pos: QPoint) -> None:
        """
        Show the context menu for customizing column visibility.

        This method shows a context menu with actions for hiding or showing columns
        based on the current visibility state.

        Args:
            pos (QPoint): The position of the context menu.

        Returns:
            None

        """
        menu = QMenu(self.__view)

        for col, i in self.get_columns().items():

            act = QAction(col, self.__view, checkable=True)
            act.setChecked(not self.__view.isColumnHidden(i))
            act.toggled.connect(partial(self._update_columns, col))
            menu.addAction(act)

        menu.exec_(self.__view.mapToGlobal(pos))

    def reset_table_state(self) -> None:
        """
        Reset the state of the table.

        This method resets the state of the table, including the column widths and visibility,
        to the initial state.

        """
        with open_settings(mode='w') as settings:
            settings.view.update({self.__parent: {'state': '', 'columns': {}}})
        self.load_table_state()



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide2.QtCore import Qt, Slot
from PySide2.QtWidgets import QMenu, QAction, QWidget

from ..utils import open_file

if TYPE_CHECKING:
    from .ui_view_model import TaskTreeView


class TaskFileMenu(QMenu):
    def __init__(self, file_path: str, parent: QWidget) -> None:
        super().__init__(title='File', parent=parent)
        self._file_path = file_path

        self._open_action = QAction('Open', self)
        self._open_action.triggered.connect(
            lambda: open_file(self._file_path)
        )
        self.addAction(self._open_action)

        self._reveal_action = QAction('Reveal', self)
        self._reveal_action.triggered.connect(
            lambda: open_file(self._file_path, reveal=True)
        )
        self.addAction(self._reveal_action)


class TaskCheckerMenu(QMenu):
    def __init__(self, parent: TaskTreeView) -> None:
        super().__init__(title='Check/Uncheck', parent=parent)

        self._parent = parent
        self.tasks_model = parent.tasks_model

        check_all = QAction('Check All', self)
        check_all.triggered.connect(lambda: self._iter_all(Qt.Checked))

        uncheck_all = QAction('Uncheck All', self)
        uncheck_all.triggered.connect(lambda: self._iter_all(Qt.Unchecked))

        invert_all = QAction('Invert All', self)
        invert_all.triggered.connect(self._invert_all)

        check_selected = QAction('Check Selected', self)
        check_selected.triggered.connect(lambda: self._iter_selected(Qt.Checked))

        uncheck_selected = QAction('Uncheck Selected', self)
        uncheck_selected.triggered.connect(lambda: self._iter_selected(Qt.Unchecked))

        self.addAction(check_all)
        self.addAction(uncheck_all)
        self.addAction(invert_all)
        self.addSeparator()
        self.addAction(check_selected)
        self.addAction(uncheck_selected)

    @Slot(Qt.CheckState)
    def _iter_all(self, state: Qt.CheckState = Qt.Checked) -> None:
        for i in range(self.tasks_model.rowCount()):
            item = self.tasks_model.item(i, 0)
            item.setCheckState(state)

    @Slot()
    def _invert_all(self) -> None:
        for i in range(self.tasks_model.rowCount()):
            item = self.tasks_model.item(i, 0)
            item.setCheckState(
                Qt.Unchecked if item.checkState() == Qt.Checked else Qt.Checked
            )

    @Slot(Qt.CheckState)
    def _iter_selected(self, state: Qt.CheckState = Qt.Checked) -> None:
        for item in self._parent.get_selected_items():
            item.setCheckState(state)



from __future__ import annotations

from typing import Tuple, Union

from PySide2.QtGui import QColor
from PySide2.QtCore import Qt

RGBA = Tuple[int, int, int, int]
TASK_COLOR = Union[str, Qt.GlobalColor, RGBA, QColor]



from __future__ import annotations

import os
import sys
from functools import cache

from PySide2.QtGui import QFont, QFontDatabase


def _load_custom_monospace_font() -> str:
    """Load bundled JetBrains Mono font from resources"""
    # Try to load from resources
    font_id = QFontDatabase.addApplicationFont(':/font/JetBrainsMonoNL-Regular.ttf')

    if font_id >= 0:
        # Successfully loaded from resources
        font_families = QFontDatabase.applicationFontFamilies(font_id)
        if font_families:
            return font_families[0]

    # Fallback to system font
    user_font = os.getenv('TQM_MONO_FONT', '')
    if user_font:
        return user_font

    if sys.platform == 'darwin':
        return 'Menlo'
    elif sys.platform == 'win32':
        return 'Consolas'
    return 'DejaVu Sans Mono'


@cache
def get_monospace_font(size: int = 12) -> QFont:
    """Get the application monospace font with specified size"""
    font_family = _load_custom_monospace_font()
    return QFont(font_family, size)



"""
Queue Task Manager: PySide2 parallel task queuing and execution framework.
"""

from __future__ import annotations

from .version import __version__

from ._core import logger
from ._core.logger import write_log
try:
    from . import _resources_rc
except ImportError:
    import warnings
    warnings.warn(
        "Resource file '_resources_rc.py' not found. "
        'Some UI resources may not be available. '
        'If you are developing, ensure resources are compiled and available in the package.',
    )


from ._main import TQManager
from . import exceptions
from ._core.task import TaskExecutable, TaskGroup, TaskUnit
from ._core.task_builder import TaskBuilder, TaskGroupBuilder
from ._core.task_executor import TaskExecutor
from ._core.task_actions import TaskActionVisibility
from .widgets.help_widget import about


__all__ = [
    'write_log',
    'exceptions',
    'TaskExecutable',
    'TaskUnit',
    'TaskGroup',
    'TaskBuilder',
    'TaskGroupBuilder',
    'TaskActionVisibility',
    'TQManager',
    'TaskExecutor',
    'about',
    '__version__',
]



from __future__ import annotations

from enum import Enum, auto
from typing import Any, Dict, Generic, TypeVar, Callable
from dataclasses import dataclass

from ..utils import extract_fn_name

T = TypeVar('T')


class TaskActionVisibility(Enum):
    ON_COMPLETED = auto()
    ON_FAILED = auto()
    ALWAYS = auto()


@dataclass(frozen=True)
class TaskAction(Generic[T]):
    name: str
    action: Callable[[T], Any]
    visibility: TaskActionVisibility = TaskActionVisibility.ON_COMPLETED

    def inspect(self) -> Dict[str, str]:
        return {
            'name': self.name,
            'action': extract_fn_name(self.action),
            'visibility': self.visibility.name,
        }




from __future__ import annotations

from random import randint
from collections import OrderedDict
from dataclasses import field, dataclass

from PySide2.QtGui import QColor


@dataclass
class RandomColor:
    """

    The class generates a unique random color each time the `generate` method is
    called. The class stores limited number of colors in the `colors` dictionary.

    Attributes:
        min_value (int): The minimum value for each color channel (red, green, blue).
        max_value (int): The maximum value for each color channel (red, green, blue).
        alpha (int): The alpha value for the color.
        max_colors (int): The maximum number of colors to generate. (Defaults 500)

    Methods:
        generate(): Generates a random color and adds it to the colors dictionary.

    Examples:
        >>> color = RandomColor().generate()
        >>> color
        <PySide2.QtGui.QColor object at 0x000001F7E9E2D9D0>
        >>> color.name()
        '#f7d2a0'

    """

    min_value: int = 20
    max_value: int = 255
    alpha: int = 255
    max_colors: int = 500

    colors: OrderedDict[str, QColor] = field(default_factory=OrderedDict[str, QColor], init=False)

    def generate(self) -> QColor:

        while True:
            r, g, b = (randint(self.min_value, self.max_value) for _ in range(3))

            color = QColor(r, g, b, self.alpha)

            if color.name() not in self.colors:

                if len(self.colors) >= self.max_colors:
                    self.colors.popitem(last=False)

                self.colors[color.name()] = color
                return color



from __future__ import annotations

from typing import Optional

from PySide2.QtWidgets import QWidget, QToolButton


class ToolButton(QToolButton):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setObjectName('TqmToolButton')



from __future__ import annotations

from typing import Any, Dict, List, Optional

from PySide2.QtGui import QFont, QColor
from PySide2.QtCore import Qt
from PySide2.QtWidgets import (QLabel, QWidget, QSplitter, QTabWidget,
                               QHeaderView, QPushButton, QTreeWidget,
                               QVBoxLayout, QTableWidget, QTreeWidgetItem,
                               QTableWidgetItem)

from tqm._core.task_executor import TaskExecutor

from ..widgets import Frame
from .._core.task import TaskUnit, TaskGroup
from .font_loader import get_monospace_font
from .._core.task_state import TaskStates


class StateHistoryWidget(QTreeWidget):

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)

        self.setColumnCount(5)
        self.setHeaderLabels(['#', 'State', 'Comment', 'Duration', 'Time'])
        self.setAlternatingRowColors(True)
        self.setStyleSheet('background-color: #282828;')
        self.setFont(get_monospace_font())
        self.header().setObjectName('TqmHeaderView')

        self.bold_font = QFont()
        self.bold_font.setBold(True)

    def populate(self, history: List[Dict[str, str]]) -> None:
        """Populate with formatted state history data."""
        self.clear()

        for i, entry in enumerate(history, 1):
            item = QTreeWidgetItem(self)

            item.setText(0, str(i))

            state = entry['state']
            item.setText(1, state.upper())
            item.setFont(1, self.bold_font)

            item.setText(2, entry['comment'])
            item.setText(3, entry['duration'])
            item.setText(4, entry['timestamp'])

            bg_color = TaskStates.get_color(state)
            for col in range(5):
                item.setForeground(col, bg_color)

        for i in range(self.columnCount()):
            self.resizeColumnToContents(i)


class TaskPropertyTreeItem(QTreeWidgetItem):
    """Tree widget item for displaying task properties."""

    def __init__(self, parent: QTreeWidgetItem, key: str, value: Any, is_parent: bool = False):
        super().__init__(parent)
        self.key = key
        self.value = value
        self.is_parent_node = is_parent
        self.setText(0, key)

        if not is_parent:
            self.setText(1, str(value))

        # Set different styling for parent nodes
        if is_parent:
            self.setForeground(0, QColor(60, 120, 216))  # Blue color for parents
            font = self.font(0)
            font.setPixelSize(14)
            font.setBold(True)
            self.setFont(0, font)


class TasksList(QWidget):
    def __init__(self, executor: TaskExecutor, parent: Optional[QWidget] = None):
        super().__init__(parent)

        self._executor = executor

        self._tasks_table = QTableWidget()
        self._tasks_table.setFont(get_monospace_font())
        self._tasks_table.setColumnCount(5)
        self._tasks_table.setEditTriggers(QHeaderView.NoEditTriggers)
        self._tasks_table.setSelectionBehavior(QHeaderView.SelectRows)
        self._tasks_table.setAlternatingRowColors(True)
        self._tasks_table.setStyleSheet('background-color: #282828;')
        self._tasks_table.setSortingEnabled(True)
        self._tasks_table.setHorizontalHeaderLabels(
            ['Name', 'Type', 'State', 'In Queue', 'Parent']
        )

        self._tasks_table.horizontalHeader().setObjectName('TqmHeaderView')
        self._tasks_table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        refresh_registry = QPushButton('Refresh')
        refresh_registry.clicked.connect(self._load_tasks)

        layout = QVBoxLayout()
        layout.addWidget(refresh_registry)
        layout.addWidget(self._tasks_table)
        self.setLayout(layout)

    def _load_tasks(self) -> None:
        tasks = self._executor.registry
        self._tasks_table.setRowCount(len(tasks))

        for row, task in enumerate(tasks):

            # name
            self._tasks_table.setItem(row, 0, QTableWidgetItem(task.name))

            # type
            obj_type = 'Group' if isinstance(task, TaskGroup) else 'Task'
            self._tasks_table.setItem(row, 1, QTableWidgetItem(obj_type))

            # state
            self._tasks_table.setItem(row, 2, QTableWidgetItem(task.state.current))

            # in queue
            in_queue = 'True' if task in self._executor.queue else 'False'
            self._tasks_table.setItem(row, 3, QTableWidgetItem(in_queue))

            # parent
            self._tasks_table.setItem(row, 4, QTableWidgetItem(str(task.parent)))


class DebugWidget(Frame):
    def __init__(self, executor: TaskExecutor, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setStyleSheet('font-size: 14px;')

        self._property_tree = QTreeWidget()
        self._property_tree.header().setObjectName('TqmHeaderView')
        self._property_tree.setFont(get_monospace_font())
        self._property_tree.setColumnCount(2)
        self._property_tree.setHeaderLabels(['Property', 'Value'])
        self._property_tree.setAlternatingRowColors(True)
        self._property_tree.setStyleSheet('background-color: #282828;')

        self._state_tree = StateHistoryWidget()

        tasks_registry = TasksList(executor)

        s = QSplitter(Qt.Vertical)
        s.addWidget(self._property_tree)
        s.addWidget(self._state_tree)

        tabs = QTabWidget()
        tabs.addTab(s, 'Task Properties')
        tabs.addTab(tasks_registry, 'Registry')

        layout = QVBoxLayout(self)
        layout.addWidget(QLabel('<h3>Debug</h3>'))
        layout.addWidget(tabs)
        self.setLayout(layout)

    def _add_properties(self, parent: QTreeWidgetItem, data: Dict[str, Any]) -> None:
        """Recursively add dictionary data to tree."""

        for key, value in sorted(data.items()):
            if isinstance(value, dict):
                item = TaskPropertyTreeItem(parent, key, '', True)
                self._add_properties(item, value)

            elif isinstance(value, list) and value and isinstance(value[0], dict):
                item = TaskPropertyTreeItem(parent, key, f"({len(value)} items)", True)
                for i, sub_value in enumerate(value):
                    child = TaskPropertyTreeItem(item, f"[{i}]", '', True)
                    if isinstance(sub_value, dict):
                        self._add_properties(child, sub_value)
                    else:
                        TaskPropertyTreeItem(child, '', str(sub_value))

            elif isinstance(value, list):
                item = TaskPropertyTreeItem(parent, key, f"({len(value)} items)", True)
                for i, sub_value in enumerate(value):
                    TaskPropertyTreeItem(item, f"[{i}]", str(sub_value))

            else:
                TaskPropertyTreeItem(parent, key, value)

    def populate(self, task: TaskUnit) -> None:
        """Populate the dialog with task data."""
        self._property_tree.clear()

        task_data = task.inspect()

        # Populate the property tree
        self._add_properties(self._property_tree.invisibleRootItem(), task_data)

        # Populate state history with specialized widget
        if 'state' in task_data and 'history' in task_data['state']:
            self._state_tree.populate(task.state.inspect()['history'])

        # Expand property tree for better visibility
        self._property_tree.expandAll()

        # Resize property tree columns
        for i in range(self._property_tree.columnCount()):
            self._property_tree.resizeColumnToContents(i)



from __future__ import annotations

from typing import Type


class TqmError(Exception):
    """Base class for exceptions in the from tqm package."""


def create_exception(name: str, base: Type[TqmError] = TqmError) -> type[TqmError]:
    """Factory function to create new exception classes."""
    return type(name, (base,), {
        '__init__': lambda self, message: super(base, self).__init__(message)
    })


TaskError = create_exception('TaskError')
TaskGroupError = create_exception('TaskGroupError')
TaskManagerError = create_exception('TaskManagerError')
TaskManagerWorkerError = create_exception('TaskManagerWorkerError')
TaskManagerWorkerGroupError = create_exception('TaskManagerWorkerGroupError')
TaskManagerWorkerTaskError = create_exception('TaskManagerWorkerTaskError')
TaskPredicateError = create_exception('TaskPredicateError')
TaskParentError = create_exception('TaskParentError')
TaskEventError = create_exception('TaskEventError')
TaskAlreadyInQueue = create_exception('TaskAlreadyInQueue')



from __future__ import annotations

import heapq
from uuid import UUID
from typing import Dict, List, Iterator

from tqm._core.task import TaskGroup, TaskExecutable
from tqm._core.task_base import TaskBase
from tqm._core.task_runner import TaskRunner, GroupRunner

from .task import TaskUnit


class DeferredTaskNotFound(Exception):
    pass


class TaskNotFoundError(Exception):
    pass


class TasksQueue:
    """Simple heap-based priority queue for tasks.

    The class also includes a deferred queue for tasks that are not ready to be
    executed. The deferred queue is a simple dictionary with the task id as the
    key and the task as the value.

    """

    DeferredTaskNotFound = DeferredTaskNotFound
    TaskNotFound = TaskNotFoundError

    def __init__(self) -> None:
        self.heap: List[TaskUnit] = []
        self.deferred: Dict[UUID, TaskUnit] = {}

    def is_empty(self) -> bool:
        """Return True if the queue is empty."""
        return not bool(self.heap)

    def peek(self) -> TaskUnit:
        """Return the task with the highest priority in the queue."""
        if not self.is_empty():
            return self.heap[0]
        raise IndexError('Queue is empty')

    def size(self) -> int:
        """Return the number of tasks in the main queue (heap)."""
        return len(self.heap)

    def size_deferred(self) -> int:
        return len(self.deferred)

    def enqueue(self, task: TaskUnit) -> None:
        """Push a task into the queue."""
        heapq.heappush(self.heap, task)

    def dequeue(self) -> TaskUnit:
        """Pop the task with the highest priority from the queue."""
        if not self.is_empty():
            return heapq.heappop(self.heap)
        raise IndexError('Queue is empty')

    def suspend(self, task: TaskUnit) -> None:
        """Defer a task to be executed later."""
        self.deferred[task.id] = task

    def is_task_deferred(self, task: TaskUnit) -> bool:
        """Check if task is in the deferred queue."""
        return task.id in self.deferred

    def remove_from_deferred(self, task: TaskUnit) -> TaskUnit:
        """Remove a task from the deferred queue.

        Raises:
            TasksQueueDeferredItemNotFound: If the task is not found in the deferred queue.
        """
        try:
            self.deferred.pop(task.id)
        except (ValueError, KeyError) as e:
            raise DeferredTaskNotFound(
                f'Item {task.name} not found in deferred queue'
            ) from e
        return task

    def main_to_deferred(self, task: TaskUnit) -> None:
        """Transfer a task from the main queue to the deferred queue."""
        self.remove_from_queue(task)
        self.suspend(task)

    def promote_to_main(self, task: TaskUnit) -> None:
        """Transfer a task from the deferred queue to the main queue."""
        self.remove_from_deferred(task)
        self.enqueue(task)

    def remove_from_queue(self, task: TaskUnit) -> TaskUnit:
        """Remove a task from the queue.

        Raises:
            TaskNotFoundError: If the task is not found in the queue.
        """
        try:
            self.heap.remove(task)
        except ValueError as e:
            raise TaskNotFoundError(f'Item {task.name} not found in queue') from e

        heapq.heapify(self.heap)
        return task

    def clear(self) -> None:
        """Clear the queue and deferred queue."""
        self.heap.clear()
        self.deferred.clear()

    def remove_task(self, task: TaskUnit) -> None:
        """Delete a task from the queue or deferred queue.

        Raises:
            TaskNotFoundError: If the task is not found in the queue or deferred queue.
        """
        try:
            self.remove_from_queue(task)
        except self.TaskNotFound:
            try:
                self.remove_from_deferred(task)
            except self.DeferredTaskNotFound as e:
                raise self.TaskNotFound from e

    def __iter__(self) -> Iterator[TaskBase[TaskExecutable, TaskRunner] | TaskBase[TaskGroup, GroupRunner]]:
        return iter([*self.heap, *self.deferred.values()])

    def __contains__(self, task: TaskUnit) -> bool:
        """
        Check if a given task is present in the queue.

        NOTE: debug only. not optimize for fast lookup

        Args:
            task (TqmTaskUnit): The task to check for presence in the queue.

        Returns:
            bool: True if the task is found in the deferred list or the heap,
                  otherwise False.
        """

        return (
            True if task.id in self.deferred
            else any(t == task for t in self.heap)
        )



from __future__ import annotations

from typing import Any, List, Optional
from functools import partial

from PySide2.QtCore import Qt, Slot, Signal, QObject
from PySide2.QtWidgets import QWidget, QInputDialog

from .._core.task import TaskUnit, TaskGroup, TaskExecutable
from .ui_view_model import TaskManagerView
from .._core.settings import Settings, open_settings
from .._core.task_executor import TaskExecutor


class _TaskButtonsController(QObject):
    """The buttons controller for the tasks widget.

    The buttons controller is responsible for handling the signals emitted by the buttons.

    Signals:
        remove_task: Signal emitted when the remove task button is clicked.
        retry_task: Signal emitted when the retry task button is clicked.
        start_workers: Signal emitted when the run all tasks button is clicked.

    """

    workers_started = Signal()
    max_workers_updated = Signal(int)

    def __init__(
        self,
        view: TaskManagerView,
        executor: TaskExecutor,
        controller: TaskManagerController,
        parent: Optional[QWidget] = None
    ) -> None:
        super().__init__(parent)

        self._view = view
        self._executor = executor
        self._controller = controller

        view.toolbar.retry_all_btn.clicked.connect(self.retry_all_failed)

        clear_tasks = self._view.toolbar.clear_tasks
        clear_tasks.completed_tasks_removed.connect(self.clear_completed_tasks)
        clear_tasks.failed_tasks_removed.connect(self.clear_failed_tasks)
        clear_tasks.waiting_tasks_removed.connect(self.clear_waiting_tasks)
        clear_tasks.all_tasks_removed.connect(self.clear_all_tasks)

        table_ops = self._view.toolbar.table_options
        table_ops.reset_layout.triggered.connect(self.reset_layout)
        table_ops.resize_columns.triggered.connect(self.resize_columns)
        table_ops.expand_all.toggled.connect(self.toggle_expand)
        table_ops.set_max_workers.triggered.connect(self.set_max_workers)

    @Slot()
    def set_max_workers(self, value: int = 20) -> None:
        """
        Sets the maximum number of workers for the application.

        If a value is provided, it emits a signal to update the maximum workers
        and optionally stores the value. If no value is provided, it retrieves
        the current maximum workers from the settings and prompts the user
        to input a new value through a dialog.

        Args:
            value (int, optional): The number of workers to set. Defaults to 20.

        Emits:
            max_workers_updated (int): Signal emitted with the new value.

        """
        if not value:

            with open_settings() as s:
                value = s.max_workers

            value, ok = QInputDialog.getInt(
                self._view, 'Set Max Workers',
                'Enter the number of workers (min 2)', value=value, minValue=2
            )
            if not ok:
                return

        value = max(2, value)
        self.max_workers_updated.emit(value)

        with open_settings('w') as s:
            s.max_workers = value

    @Slot()
    def toggle_expand(self, expand: bool) -> None:
        self._view.toggle_expand(expand)

    @Slot()
    def reset_layout(self) -> None:
        self._view.tree_view.reset_table_state()

    @Slot()
    def resize_columns(self) -> None:
        for i in range(self._view.tree_view.tasks_model.columnCount()):
            self._view.tree_view.setColumnWidth(i, 250)

    @Slot()
    def retry_all_failed(self) -> None:
        for task in self._controller.get_all_tasks():
            if task.state.is_failed:
                self._executor.retry_task(task)

    @Slot()
    def clear_completed_tasks(self) -> None:
        for task in self._controller.get_all_tasks():
            if task.state.is_completed:
                self._executor.remove_task(task)

    @Slot()
    def clear_failed_tasks(self) -> None:
        for task in self._controller.get_all_tasks():
            if task.state.is_failed:
                self._executor.remove_task(task)

    @Slot()
    def clear_waiting_tasks(self) -> None:
        for task in self._controller.get_all_tasks():
            if task.state.is_waiting:
                self._executor.remove_task(task)

    @Slot()
    def clear_all_tasks(self) -> None:
        for task in self._controller.get_all_tasks():
            self._executor.remove_task(task)


class TaskManagerController(QObject):
    task_removed = Signal(object)
    task_retried = Signal(object)

    def __init__(
        self,
        view: TaskManagerView,
        executor: TaskExecutor,
        settings: Settings,
        parent: Optional[QObject] = None
    ) -> None:
        super().__init__(parent)

        self.view = view
        self._settings = settings
        self._executor = executor

        view.toolbar.run_all_tasks.clicked.connect(executor.start_workers)

        executor.callbacks.task_added.connect(self.add_task)
        executor.callbacks.runner_completed.connect(self._on_task_completed)
        executor.callbacks.runner_started.connect(self._on_task_started)
        executor.callbacks.runner_finished.connect(self._on_task_finished)
        executor.callbacks.task_removed.connect(self.remove_task)

        self.ops = _TaskButtonsController(self.view, executor, self)

        self.view.tree_view.selected_task_removed.connect(self._on_remove_selected_tasks)
        self.view.tree_view.selected_task_retried.connect(self._on_retry_selected)

    def _update_item_data(
        self,
        task: TaskUnit,
        column_name: str,
        value: Any,
        role: Qt.ItemDataRole = Qt.DisplayRole
    ) -> None:
        """Set data to a task item."""
        if not task.item:
            return

        parent = task.item.parent()
        item = parent.child if parent else self.view.tree_view.tasks_model.item
        column = self.view.tree_view.get_column_index(column_name)
        item(task.item.row(), column).setData(value, role)

    @Slot(object, float)
    def _on_task_update_progress(self, task: TaskUnit, value: float) -> None:
        """Update the progress bar of a task."""
        self._update_item_data(task, 'Progress', round(value, 2))

    @Slot(object)
    def _on_task_completed(self, task: TaskExecutable, *args: Any, **kwargs: Any) -> None:
        self._on_task_update_progress(task, task.progress_bar.maximum)
        self._update_item_data(task, 'Completed', task.state.get_last().timestamp)

    @Slot(object)
    def _on_task_finished(self, task: TaskUnit) -> None:
        if not isinstance(task, TaskExecutable):
            return

        group = task.group
        if not group:
            return

        completed_tasks = len(list(filter(lambda t: t.state.is_completed, group.tasks)))
        self._update_item_data(group, 'Progress', completed_tasks)

    @Slot(object)
    def _on_task_started(self, task: TaskUnit) -> None:
        self._update_item_data(task, 'Started', task.state.get_last().timestamp)

    @Slot()
    def _on_retry_selected(self) -> None:
        for task in self.get_selected_tasks():
            self._executor.retry_task(task)

    def add_task(self, task: TaskUnit) -> None:
        task.item = self.view.tree_view.tasks_model.add_task(task)
        task.runner.signals.task_progress_updated.connect(
            partial(self._on_task_update_progress, task)
        )
        self.view.toggle_expand(True)

    def remove_task(self, task: TaskUnit) -> None:
        """
        Removes a task from the tasks model and its associated tree view.

        Args:
            task (TqmTaskUnit): The task to be removed. It must have an associated
                item in the tree view.

        Behavior:
            - If the task has no associated item (`task.item` is None), the method
              returns immediately without performing any action.
            - If the task's item has a parent group (`group_parent`):
                - The task's item is removed from its parent group.
                - If the parent group becomes empty after the removal, the parent
                  group itself is removed from the tasks model.
            - If the task's item does not have a parent group, it is removed
              directly from the tasks model.
        """
        if not task.item:
            return

        remove_row = task.item.row()

        group_parent = task.item.parent()

        if group_parent:
            group_parent.removeRow(remove_row)

            # remove empty groups
            if not group_parent.rowCount():
                self._executor.remove_task(group_parent.data(Qt.UserRole))
        else:
            self.view.tree_view.tasks_model.removeRow(remove_row)

    @Slot()
    def _on_remove_selected_tasks(self) -> None:
        for task in self.get_selected_tasks():
            self._executor.remove_task(task)

    def get_all_tasks(self) -> List[TaskUnit]:
        """Get a list of all tasks in the tree view."""
        tasks: List[TaskUnit] = []
        for i in range(self.view.tree_view.tasks_model.rowCount()):
            item = self.view.tree_view.tasks_model.item(i, 0).data(Qt.UserRole)
            tasks.append(item)
            if isinstance(item, TaskGroup):
                tasks.extend(item.tasks)
        return tasks

    def get_selected_tasks(self) -> List[TaskUnit]:
        """Get a list of selected tasks in the tree view."""
        return [item.data(Qt.UserRole) for item in self.view.tree_view.get_selected_items()]



from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Tuple, Callable, Optional
from datetime import datetime
from dataclasses import field, dataclass

from PySide2.QtGui import QColor


class TaskStateEnum(str, Enum):
    """Enum representing the possible states of a task."""
    RUNNING = 'running'
    COMPLETED = 'completed'
    FAILED = 'failed'
    BLOCKED = 'blocked'
    RETRYING = 'retrying'
    DELETED = 'deleted'
    WAITING = 'waiting'
    INACTIVE = 'inactive'


class TaskStates:
    """Singleton class managing the configuration for task states."""

    COLORS = {
        TaskStateEnum.RUNNING:   QColor('#61AFEF'),  # BLUE
        TaskStateEnum.COMPLETED: QColor('#3FC13F'),  # GREEN
        TaskStateEnum.BLOCKED:   QColor('#D19A66'),  # ORANGE
        TaskStateEnum.RETRYING:  QColor('#C678DD'),  # PURPLE
        TaskStateEnum.FAILED:    QColor('#E06C75'),  # BRIGHT RED
        TaskStateEnum.DELETED:   QColor('#BE5046'),  # DARKER RED
        TaskStateEnum.WAITING:   QColor('#8A8D94'),  # LIGHT GREY
        TaskStateEnum.INACTIVE:  QColor('#5C6370'),  # MEDIUM GREY
    }

    @classmethod
    def get_color(cls, state: TaskStateEnum) -> QColor:
        """Get the color for a task state."""
        return cls.COLORS[state]


def format_duration(seconds: float) -> str:
    """Format duration in a human-readable form."""
    if seconds < 0.1:
        return f"+{int(seconds * 1000)}ms"
    elif seconds < 10:
        return f"+{seconds:.3f}s"
    elif seconds < 60:
        return f"+{int(seconds)}s"
    elif seconds < 3600:
        minutes = int(seconds // 60)
        remaining_seconds = int(seconds % 60)
        return f"+{minutes}m {remaining_seconds}s"

    hours = int(seconds // 3600)
    minutes = int((seconds % 3600) // 60)
    return f"+{hours}h {minutes}m"


@dataclass
class StateHistory:
    active_state: str
    comment: str = ''
    timestamp: str = field(init=False)

    def __post_init__(self) -> None:
        self.timestamp = datetime.now().isoformat()


@dataclass
class TaskState:
    """TaskState.

    This class represents the state of a task. A task always starts in the inactive state.
    """
    current: TaskStateEnum = TaskStateEnum.INACTIVE
    history:  Tuple[StateHistory, ...] = field(default_factory=tuple)

    def __post_init__(self) -> None:
        self._on_state_changed: Optional[Callable[[str, str], None]] = None

    def __str__(self) -> str:
        return self.current.name.lower()

    @property
    def color(self) -> QColor:
        """Get the color associated with the current state."""
        return TaskStates.get_color(self.current)

    def inspect(self) -> Dict[str, Any]:
        return {
            'current': self.current.value,
            'history': self._process_history_entries()
        }

    def _process_history_entries(self) -> List[Dict[str, Any]]:
        """Process history entries to add duration and formatting."""

        result: List[Dict[str, Any]] = []
        prev_time: Optional[datetime] = None

        for entry in self.history:
            timestamp = entry.timestamp
            current_time = datetime.fromisoformat(timestamp.replace('Z', '+00:00'))

            # Calculate duration since previous state
            duration = ''
            if prev_time is not None:
                diff = current_time - prev_time
                duration = format_duration(diff.total_seconds())

            result.append({
                'state': entry.active_state,
                'comment': entry.comment,
                'duration': duration,
                'timestamp': timestamp
            })

            prev_time = current_time

        return result

    def _set_state(self, state: TaskStateEnum, comment: str = '') -> None:
        previous_state = self.current
        self.current = state
        history = StateHistory(self.current.value, comment)
        self.history = self.history + (history,)

        if self._on_state_changed:
            self._on_state_changed(previous_state, state)

    def register_state_change_callback(
        self,
        on_state_changed: Callable[[str, str], Any]
    ) -> None:
        self._on_state_changed = on_state_changed

    def set_running(self, comment: str = '') -> None:
        self._set_state(TaskStateEnum.RUNNING, comment)

    def set_completed(self, comment: str = '') -> None:
        self._set_state(TaskStateEnum.COMPLETED, comment)

    def set_failed(self, comment: str = '') -> None:
        self._set_state(TaskStateEnum.FAILED, comment)

    def set_blocked(self, comment: str = '') -> None:
        self._set_state(TaskStateEnum.BLOCKED, comment)

    def set_retrying(self, comment: str = '') -> None:
        self._set_state(TaskStateEnum.RETRYING, comment)

    def set_deleted(self, comment: str = '') -> None:
        self._set_state(TaskStateEnum.DELETED, comment)

    def set_waiting(self, comment: str = '') -> None:
        self._set_state(TaskStateEnum.WAITING, comment)

    def set_inactive(self, comment: str = '') -> None:
        self._set_state(TaskStateEnum.INACTIVE, comment)

    def get_first(self) -> StateHistory:
        return self.history[0]

    def get_last(self) -> StateHistory:
        return self.history[-1]

    @property
    def is_running(self) -> bool:
        return self.current == TaskStateEnum.RUNNING

    @property
    def is_completed(self) -> bool:
        return self.current == TaskStateEnum.COMPLETED

    @property
    def is_failed(self) -> bool:
        return self.current == TaskStateEnum.FAILED

    @property
    def is_blocked(self) -> bool:
        return self.current == TaskStateEnum.BLOCKED

    @property
    def is_retrying(self) -> bool:
        return self.current == TaskStateEnum.RETRYING

    @property
    def is_deleted(self) -> bool:
        return self.current == TaskStateEnum.DELETED

    @property
    def is_waiting(self) -> bool:
        return self.current == TaskStateEnum.WAITING

    @property
    def is_inactive(self) -> bool:
        return self.current == TaskStateEnum.INACTIVE

    @property
    def is_active(self) -> bool:
        return any([
            self.is_running, self.is_retrying, self.is_blocked,
        ])

    @property
    def is_removable(self) -> bool:
        return any([
            self.is_waiting, self.is_deleted, self.is_inactive,
            self.is_completed, self.is_failed
        ])




from __future__ import annotations

from enum import Enum, auto
from typing import Any, Dict, Callable, Optional
from dataclasses import field, dataclass

from PySide2.QtCore import QTimer

from ..utils import extract_fn_name


class PredicateEventType(Enum):
    START = auto()
    RETRY = auto()
    SUCCESS = auto()
    FAIL = auto()


@dataclass
class TaskPredicate:
    condition: Optional[Callable[..., bool]] = None
    max_retries: int = 2
    retry_interval: int = 1000

    timer: QTimer = field(init=False, repr=False)

    def __post_init__(self) -> None:
        self.timer = QTimer()
        self.timer.setInterval(self.retry_interval)
        self.retry_left = self.max_retries

    def reset(self) -> None:
        self.retry_left = self.max_retries

    def delete(self) -> None:
        if self.timer.isActive():
            self.timer.stop()
            self.timer.deleteLater()
        self.condition = None

    def stop_timer(self) -> None:
        if self.timer.isActive():
            self.timer.stop()
            self.timer.timeout.disconnect()

    def evaluate(self, predicate_handler: Callable[[PredicateEventType], Any]) -> None:
        """Evaluate the predicate condition with retry logic.

        This method checks if the predicate condition is met. If not, it will retry
        according to the configured retry settings. The status_notifier callback
        is called with different PredicateEventType values to report the state.

        Args:
            status_notifier: Callback function that receives status updates

        NOTE: See TaskExecutor._handle_predicate for an example
        """
        condition = self.condition
        if condition is None:
            return

        if self.retry_left <= 0:
            predicate_handler(PredicateEventType.FAIL)
            return

        def _evaluate() -> None:
            self.retry_left -= 1
            predicate_handler(PredicateEventType.RETRY)

            if condition():
                self.delete()
                predicate_handler(PredicateEventType.SUCCESS)

            elif self.retry_left <= 0:
                self.stop_timer()
                predicate_handler(PredicateEventType.FAIL)

        if not self.timer.isActive():
            self.timer.timeout.connect(_evaluate)
            self.timer.start()

    def inspect(self) -> Dict[str, Any]:
        return {
            'fn': extract_fn_name(self.condition),
            'retries': self.max_retries,
            'interval': self.retry_interval
        }



from __future__ import annotations

from typing import Any, Set, Dict, Union, Callable, Optional
from dataclasses import field, dataclass

from ..utils import extract_fn_name
from .logger import LOGGER
from .task_base import TaskBase
from .task_runner import TaskRunner, GroupRunner

TaskUnit = Union[
    TaskBase['TaskExecutable', TaskRunner],
    TaskBase['TaskGroup', GroupRunner]
]


@dataclass
class TaskExecutable(TaskBase['TaskExecutable', TaskRunner]):
    execute: Callable[..., Any] = lambda: None
    group: Optional[TaskGroup] = None
    runner: TaskRunner = field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__(suffix='Task')
        LOGGER.debug('TaskExecutable created: %s', str(self))
        self.runner = TaskRunner(self)

    def __hash__(self) -> int:
        return hash(self.id)

    def reset(self, comment: str = '', reset_attempts: bool = False) -> None:
        super().reset(comment, reset_attempts)
        self.runner = TaskRunner(self)

    def delete(self, comment: str = '') -> None:
        if self.group:
            self.group.tasks.remove(self)
        return super().delete(comment)

    def emit_progress(self, value: float) -> None:
        self.runner.signals.task_progress_updated.emit(value)

    def inspect(self) -> Dict[str, Any]:
        data = super().inspect()
        data['execute'] = extract_fn_name(self.execute)
        data['group'] = str(self.group) or ''
        return data


@dataclass
class TaskGroup(TaskBase['TaskGroup', GroupRunner]):
    tasks: Set[TaskExecutable] = field(init=False, default_factory=set[TaskExecutable])
    runner: GroupRunner = field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__(suffix='Group')
        LOGGER.debug('Task Group Created: %s', str(self))
        self.runner = GroupRunner(self)

    def __hash__(self) -> int:
        return hash(self.id)

    def reset(self, comment: str = '', reset_attempts: bool = False) -> None:
        super().reset(comment, reset_attempts)
        self.runner = GroupRunner(self)

    def add_tasks(self, *tasks: TaskExecutable) -> None:
        for task in tasks:
            self.tasks.add(task)
            task.group = self

    def add_event(
        self,
        execute: Callable[[TaskExecutable], Any],
        *,
        show_progress: bool = False,
        label: str = '',
        comment: str = ''
    ) -> TaskExecutable:
        """Add a simple task to the queue.

        Args:
            execute (Callable[..., Any]): The event to be executed. The first argument is the task itself.
            progress (bool): If the task should show a progress.
            label (Optional[str], optional): The label of the task. If not provided,
                a default label will be generated.
            details (Optional[str], optional): Comment of the task. Defaults to ''.

        ```
        add_event(
           execute=lambda task: print(task.name),
           progress=True,
           label='My task',
           comment='Comment of the task'
        )
        ```
        """
        # XXX: find a better way to avoid circular dependency
        from .task_builder import TaskBuilder
        task = (
            TaskBuilder(name=label)
            .with_event(execute, show_progress=show_progress)
            .with_comment(comment)
            .build()
        )
        self.add_tasks(task)
        return task

    def inspect(self) -> Dict[str, Any]:
        data = super().inspect()
        data['tasks'] = [str(task) for task in self.tasks]
        return data



from __future__ import annotations

import os
import sysconfig
from typing import Any, Dict, Optional
from platform import python_version

from PySide2 import __version__ as PySide2_version
from PySide2.QtGui import QDesktopServices
from PySide2.QtCore import QUrl, Slot
from PySide2.QtWidgets import (QLabel, QDialog, QWidget, QGridLayout,
                               QPushButton, QVBoxLayout, QPlainTextEdit)

from ..version import __version__


class HelpWidget(QDialog):
    def __init__(
        self,
        git_repo: str,
        config_path: str,
        about: Dict[str, Any],
        extra: Optional[Dict[str, Any]] = None,
        parent: Optional[QWidget] = None
    ) -> None:
        super().__init__(parent)
        self.setWindowTitle('Help')

        self.git_repo = git_repo
        self.config_path = config_path

        about_info = QPlainTextEdit()
        about_info.setReadOnly(True)

        extra = extra or {}
        abouts = {**about, **extra}

        for name, value in abouts.items():
            about_info.appendPlainText(f'- {name.title()}: {value}')

        grid_layout = QGridLayout()
        grid_layout.addWidget(self._button_factory('Issues'), 0, 0, 1, 2)
        grid_layout.addWidget(self._button_factory('Readme'), 1, 0)
        grid_layout.addWidget(self._button_factory('Changelog'), 1, 1)
        grid_layout.addWidget(self._button_factory('Config'), 2, 0, 1, 2)

        layout = QVBoxLayout()
        layout.addWidget(QLabel('<h1>Help</h1>'))
        layout.addWidget(about_info)
        layout.addLayout(grid_layout)

        self.setLayout(layout)

    def _button_factory(self, text: str) -> QPushButton:
        button = QPushButton(text)
        button.clicked.connect(lambda: self._on_open_link(text))
        return button

    @Slot(str)
    def _on_open_link(self, link: str) -> None:
        links = {
            'issues': QUrl(f'{self.git_repo}/issues'),
            'changelog': QUrl(f'{self.git_repo}/blob/master/CHANGELOG.md'),
            'readme': QUrl(f'{self.git_repo}/blob/master/README.md'),
            'config': QUrl.fromLocalFile(self.config_path)
        }
        QDesktopServices.openUrl(links[link.lower()])


def about() -> Dict[str, str]:
    """Return a dictionary with information about the application."""
    return {
        'version': __version__,
        'python': python_version(),
        'pyside': PySide2_version,
        'machine': sysconfig.get_platform(),
    }


def show_help(**kwargs: Any) -> None:
    return HelpWidget(
        git_repo='https://github.com/sisoe24/_TODO_',
        config_path=os.environ['TQM_CONFIG_PATH'],
        about=about(),
        extra=kwargs
    ).show()



from __future__ import annotations

from enum import Enum
from typing import Any, Dict, Generic, TypeVar, Callable, Optional
from dataclasses import dataclass

from tqm.utils import extract_fn_name

T = TypeVar('T')


class TaskEventCallbacks(str, Enum):
    ON_START = 'on_start'
    ON_COMPLETED = 'on_completed'
    ON_FAILED = 'on_failed'
    ON_FINISH = 'on_finish'


@dataclass
class TaskCallbacks(Generic[T]):
    on_start: Optional[Callable[[T], Any]] = None
    on_finish: Optional[Callable[[T], Any]] = None
    on_failed: Optional[Callable[[T], Any]] = None
    on_completed: Optional[Callable[[T], Any]] = None

    def delete(self) -> None:
        self.on_start = None
        self.on_finish = None
        self.on_failed = None
        self.on_completed = None

    def inspect(self) -> Dict[str, str]:
        return {
            'on_start': extract_fn_name(self.on_start),
            'on_finish': extract_fn_name(self.on_finish),
            'on_failed': extract_fn_name(self.on_failed),
            'on_completed': extract_fn_name(self.on_completed),
        }



from __future__ import annotations

import random
from uuid import UUID
from typing import TYPE_CHECKING, Dict

from PySide2.QtGui import QPainter, QPalette
from PySide2.QtCore import Qt, QTimer, QObject, QModelIndex
from PySide2.QtWidgets import (QStyle, QApplication, QStyledItemDelegate,
                               QStyleOptionViewItem, QStyleOptionProgressBar)

from .._core.task import TaskUnit, TaskGroup
from .._core.task_options import ProgressMode

if TYPE_CHECKING:
    from .ui_view_model import TaskTreeView


class ProgressBarAnimator(QObject):
    """Handles animation state for progress bars"""

    def __init__(self, parent: TaskTreeView) -> None:
        super().__init__(parent)
        self._parent = parent

        self._animation_counter = 0
        self._task_offsets: Dict[UUID, int] = {}

        self._animation_timer = QTimer(self)
        self._animation_timer.setInterval(50)  # 50ms = 20fps
        self._animation_timer.timeout.connect(self._update_animation)

    def start_animation(self) -> None:
        """Start the animation timer if not already running"""
        if not self._animation_timer.isActive():
            self._animation_timer.start()

    def stop_animation(self) -> None:
        """Stop the animation timer if running"""
        if self._animation_timer.isActive():
            self._animation_timer.stop()

    def _update_animation(self) -> None:
        """Update the animation counter on timer tick."""
        self._animation_counter = (self._animation_counter + 3) % 200
        self._parent.viewport().update()

    def get_task_animation_value(self, task: TaskUnit) -> int:
        """Get animation value for a specific task, with random offset for variety."""

        # Generate a random offset for this task if it doesn't exist
        if task.id not in self._task_offsets:
            self._task_offsets[task.id] = random.randint(0, 100)

        animation_value = (self._animation_counter + self._task_offsets[task.id]) % 200
        if animation_value > 100:
            animation_value = 200 - animation_value  # reverse direction after 100

        return animation_value

    def cleanup(self) -> None:
        """Stop timer and clear resources"""
        self.stop_animation()
        self._task_offsets.clear()


class ProgressBarRenderer:
    """Handles rendering logic for progress bars"""

    @staticmethod
    def handle_determinate_progress(
        pbar_options: QStyleOptionProgressBar,
        progress: int,
        task: TaskUnit
    ) -> None:
        """Configure progress bar options for determinate progress"""
        pbar_options.progress = progress

        if isinstance(task, TaskGroup):
            text = f"{progress}/{len(task.tasks)}"
        else:
            text = f"{progress}%"

        pbar_options.text = text

    @staticmethod
    def handle_indeterminate_progress(
        pbar_options: QStyleOptionProgressBar,
        animation_value: int,
        task: TaskUnit
    ) -> None:
        """Configure progress bar options for indeterminate progress"""
        pbar_options.minimum = 0
        pbar_options.maximum = 100
        pbar_options.progress = animation_value
        pbar_options.text = task.progress_bar.working_text

    @staticmethod
    def handle_completed_state(
        pbar_options: QStyleOptionProgressBar,
        task: TaskUnit,
        progress: int
    ) -> None:
        """Configure progress bar options for completed/inactive tasks"""
        text = task.state.current.title()
        color = task.state.color

        if isinstance(task, TaskGroup):
            text = f'Group: {text} - {progress}/{len(task.tasks)}'
            color = color.darker()

        if task.state.is_completed:
            pbar_options.maximum = task.progress_bar.maximum

        pbar_options.palette.setColor(QPalette.Base, color)
        pbar_options.text = text


class ProgressBarDelegate(QStyledItemDelegate):
    """A delegate that renders progress bars for tasks"""

    def __init__(self, progress_column: int, parent: TaskTreeView) -> None:
        super().__init__(parent)
        self.progress_column = progress_column

        self._animator = ProgressBarAnimator(parent)
        self._renderer = ProgressBarRenderer()

    def start_animation(self) -> None:
        """Start the animation timer"""
        self._animator.start_animation()

    def stop_animation(self) -> None:
        """Stop the animation timer"""
        self._animator.stop_animation()

    def paint_progress_bar(
        self,
        painter: QPainter,
        option: QStyleOptionViewItem,
        index: QModelIndex,
        task: TaskUnit
    ) -> None:
        """Paint the progress bar for a task"""
        pbar_options = QStyleOptionProgressBar()
        pbar_options.rect = option.rect
        pbar_options.textVisible = True

        pbar_options.minimum = task.progress_bar.minimum
        pbar_options.maximum = task.progress_bar.maximum

        progress = index.data()

        if task.state.is_running:
            if task.progress_bar.mode == ProgressMode.DETERMINATE:
                self._renderer.handle_determinate_progress(pbar_options, progress, task)
            else:
                animation_value = self._animator.get_task_animation_value(task)
                self._renderer.handle_indeterminate_progress(pbar_options, animation_value, task)
                # Start animation if not already running
                self._animator.start_animation()
        else:
            self._renderer.handle_completed_state(pbar_options, task, progress)

        QApplication.style().drawControl(QStyle.CE_ProgressBar, pbar_options, painter)

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex) -> None:
        task: TaskUnit = index.siblingAtColumn(0).data(Qt.UserRole)

        if index.column() == self.progress_column:
            self.paint_progress_bar(painter, option, index, task)
            return

        super().paint(painter, option, index)

    def cleanup(self) -> None:
        """Clean up resources and stop the animation timer"""
        self._animator.cleanup()



from __future__ import annotations

from enum import Enum, auto
from typing import Any, Dict
from dataclasses import asdict, dataclass


class ProgressMode(Enum):
    DETERMINATE = auto()
    INDETERMINATE = auto()


@dataclass
class ProgressBarOptions:

    working_text: str = 'Working...'

    minimum: int = 0
    maximum: int = 100

    mode: ProgressMode = ProgressMode.DETERMINATE

    def inspect(self) -> Dict[str, Any]:
        return asdict(self)



from __future__ import annotations

import os
import sys
import logging
from typing import TYPE_CHECKING, TextIO, Optional
from pathlib import Path
from logging.handlers import TimedRotatingFileHandler

from PySide2.QtCore import Qt, Signal, QObject

if TYPE_CHECKING:
    from .._ui.tab_logs import TasksLog

USER_LEVEL = 25
logging.addLevelName(USER_LEVEL, 'USER')

LOG_FILE_FORMAT = logging.Formatter(' - '.join([
    '%(asctime)s',
    '%(levelname)-8s',
    '%(threadName)s ID:%(thread)-15d',
    '%(message)s'
]), '%Y-%m-%d %H:%M:%S')

LOG_WIDGET_FORMAT = logging.Formatter(' - '.join([
    '%(asctime)s',
    '%(levelname)-8s',
    '%(threadName)s ID:%(thread)-15d',
    '%(message)s'
]), '%H:%M:%S')


class TqmLogger(logging.Logger):
    def __init__(self, name: str = 'tqm') -> None:
        super().__init__(name)
        self.setLevel(logging.DEBUG)
        self._widget = logging.NullHandler()

    @property
    def widget(self) -> logging.Handler:
        return self._widget

    @widget.setter
    def widget(self, handler: logging.Handler) -> None:
        self._widget = handler
        self.addHandler(self._widget)


class HandlerSignal(QObject):
    log_message = Signal(str, str, str)


class WidgetLogHandler(logging.Handler):

    def __init__(self, widget: TasksLog, debug: bool = False) -> None:
        super().__init__()
        self.set_name('widget_handler')
        self.setLevel(logging.DEBUG if debug else logging.INFO)
        self.setFormatter(LOG_WIDGET_FORMAT)

        self._widget = widget
        self._signal = HandlerSignal()
        self._signal.log_message.connect(self._widget.log, Qt.QueuedConnection)

    def emit(self, record: logging.LogRecord) -> None:
        self._signal.log_message.emit(
            self.format(record) + '\n',
            record.levelname.upper(),
            str(record.thread)
        )


def log_file_handler(config_path: Path) -> TimedRotatingFileHandler:
    tqm_log_dir = config_path / 'logs'
    tqm_log_dir.mkdir(parents=True, exist_ok=True)

    handler = TimedRotatingFileHandler(
        filename=tqm_log_dir / 'tqm.log',
        when='midnight',
        backupCount=7
    )

    handler.setLevel(logging.DEBUG)
    handler.set_name('file_handler')
    handler.setFormatter(LOG_FILE_FORMAT)

    return handler


def _console_handler() -> logging.StreamHandler[TextIO]:
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.set_name('console_handler')
    handler.setFormatter(LOG_WIDGET_FORMAT)
    return handler


def write_log(msg: str, *, stream: Optional[TextIO] = None) -> None:
    """Write a log level USER message to the logger.

    Args:
        msg (str): The message to log.
        stream (Optional[TextIO], optional): The stream to write to. Defaults to None.

    ```
    write_log('Hello, World!')
    write_log('Hello, World!', stream=sys.stdout)
    ```

    """
    if stream:
        try:
            stream.write(msg + '\n')
            stream.flush()
        except Exception as e:
            LOGGER.error('Invalid stream: %s', e)

    LOGGER.log(USER_LEVEL, msg)


LOGGER = TqmLogger()
if os.getenv('TQM_DEBUG') == '1':
    LOGGER.addHandler(_console_handler())




from __future__ import annotations

import threading
from pprint import pformat
from typing import Any, Dict, Union

from PySide2.QtCore import QThread


def get_thread_info(pretty: bool = False) -> Union[str, Dict[str, Any]]:
    """Get information about the current thread."""
    current_thread = threading.current_thread()
    qt_thread = QThread.currentThread()
    info: Dict[str, Any] = {
        'name': current_thread.name,
        'id': threading.get_ident(),
        'native_id': current_thread.native_id,
        'qt_thread_id': qt_thread.currentThread()
    }
    return pformat(info) if pretty else info



from __future__ import annotations

from typing import Any, Callable, Optional, Generator
from contextlib import contextmanager

from PySide2.QtWidgets import QWidget, QTabWidget, QMainWindow, QMessageBox

from ._core.task import TaskUnit, TaskGroup, TaskExecutable
from ._core.logger import LOGGER, WidgetLogHandler, log_file_handler
from ._ui.tab_logs import TasksLog
from ._core.settings import get_qss_path, open_settings, get_config_path
from ._ui.ui_controller import TaskManagerController, _TaskButtonsController
from ._ui.ui_view_model import TaskManagerView
from ._core.task_builder import TaskBuilder, TaskGroupBuilder
from ._core.task_executor import TaskExecutor, _ExecutorCallbacks

TaskGroupGenerator = Generator[TaskGroup, Any, None]


class TQManager(QMainWindow):
    """TQManager.

    This class is the main entry point for the TQManager. It provides a simple
    interface for adding tasks to a queue and executing them.

    >>> widget = TQManager()
    >>> with widget.begin_task_insert() as tasks:
    ...    tasks.add_event(lambda t: None, label='Dummy Task', comment='Does nothing')
    >>> widget.start_workers()
    """

    def __init__(self, app_name: str, parent: Optional[QWidget] = None) -> None:
        """Initialize the TQManager.

        Args:
            app_name (str): The name of the application. This is most useful when
            running multiple instances of the TQManager so differentiate between
            them.
        """
        super().__init__(parent)

        config_path = get_config_path(app_name)
        config_path.mkdir(exist_ok=True, parents=True)

        self.setStyleSheet(get_qss_path(config_path).read_text())

        with open_settings() as s:
            settings = s

        # logs
        self._logs = TasksLog()
        LOGGER.widget = WidgetLogHandler(self._logs, settings.enable_debug)
        LOGGER.info('Initializing TQManager')
        LOGGER.addHandler(log_file_handler(config_path))

        # core logic
        self.executor = TaskExecutor(settings.max_workers)

        # widgets logic
        self._view = TaskManagerView(self.executor, parent)
        self.executor.callbacks.system_idle.connect(self._view.tree_view.stop_animation)

        self.executor.callbacks.status_updated.connect(self._view.update_status)

        self._controller = TaskManagerController(self._view, self.executor, settings)
        self._controller.ops.max_workers_updated.connect(self.executor.set_max_workers)

        self.tabs = QTabWidget()
        self.tabs.setObjectName('TqmTabsWidget')
        self.tabs.addTab(self._view, 'Tasks')
        self.tabs.addTab(self._logs, 'Logs')
        self.setCentralWidget(self.tabs)

    @property
    def callbacks(self) -> _ExecutorCallbacks:
        return self.executor.callbacks

    @property
    def operations(self) -> _TaskButtonsController:
        return self._controller.ops

    def status_message(self, text: str, timeout: int = 5000) -> None:
        self.statusBar().showMessage(text, timeout)

    def start_workers(self) -> None:
        """Start the workers."""
        self.executor.start_workers()

    def set_max_workers(self, workers: int) -> None:
        """Set the maximum number of worker threads for the executor."""
        self.executor.set_max_workers(workers)

    def remove_tasks(self, *tasks: TaskUnit) -> None:
        for task in tasks:
            self.executor.remove_task(task)

    def add_tasks(self, *tasks: TaskUnit) -> None:
        """Add a group of tasks to the queue.

        ```
        task = TaskBuilder().build()
        group = TaskGroupBuilder().build()
        task_manager.add(group, task)
        ```

        """
        for task in tasks:
            self.executor.add_task(task)
            self.callbacks.task_added.emit(task)

    def add_event(
        self,
        execute: Callable[[TaskExecutable], Any],
        *,
        show_progress: bool = False,
        label: str = '',
        comment: str = ''
    ) -> TaskExecutable:
        """Create a simple task.

        Args:
            execute (Callable[[Task], Any]): The event to be executed. The first argument
            is the task itself.
            show_progress (bool): If the task should show a progress.
            label (Optional[str], optional): The label of the task. If not provided,
                a default label will be generated.
            details (Optional[str], optional): Comment of the task. Defaults to ''.

        ```
        add_event(
           execute=lambda task: print(task.name),
           progress=True,
           label='My task',
           comment='Comment of the task'
        )
        ```
        """
        task = (
            TaskBuilder(name=label)
            .with_event(execute, show_progress=show_progress)
            .with_comment(comment)
            .build()
        )
        self.add_tasks(task)
        return task

    @contextmanager
    def create_group(self, group_name: str) -> TaskGroupGenerator:
        """Add a simple group to the queue.

        ```
        with create_group('My Group') as group:
            group.add_tasks(task1, task2)
            group.add_event(lambda t: t.log('hello'))
        ```
        """
        group = TaskGroupBuilder(name=group_name).build()
        yield group
        self.add_tasks(group)

    def shutdown(self) -> None:
        """Shutdown the workers."""
        if QMessageBox.question(
            None,
            'Shutdown',
            'Are you sure you want to shutdown the workers? '
            'Note: In-progress tasks will not be terminated.',
            QMessageBox.Yes | QMessageBox.No
        ) != QMessageBox.Yes:
            return

        self.executor.shutdown()



from __future__ import annotations

import os
import sys
import json
from typing import Any, Dict, Literal, Optional, Generator
from pathlib import Path
from contextlib import contextmanager
from dataclasses import field, asdict, dataclass


def get_config_path(app_name: str) -> Path:
    """
    Get the config directory path for the application.

    Args:
        app_name (str): The name of the application

    Returns:
        Path: The config directory path for the application.
    """
    home = Path.home()

    if sys.platform == 'win32':
        env = os.getenv('LOCALAPPDATA')
        path = (
            Path(env) / 'tqm' / app_name
            if env
            else home / 'AppData' / 'Local' / 'tqm' / app_name
        )
    else:
        env = os.environ.get('XDG_CONFIG_HOME')
        path = (
            Path(env) / 'tqm' / app_name
            if env
            else home / '.config' / 'tqm' / app_name
        )

    os.environ['TQM_CONFIG_PATH'] = str(path)
    os.environ['TQM_SETTINGS_PATH'] = str(path / 'settings.json')

    return path


def get_qss_path(config_path: Path) -> Path:
    """Returns the style qss path.

    Uses the environment variable `TQM_QSS_PATH` if set.

    """
    env = os.getenv('TQM_QSS_PATH')
    if env:
        return Path(env)

    user_qss = config_path / 'style.qss'
    base_qss = Path(__file__).parent.parent / 'theme' / 'style.qss'

    if not user_qss.exists():
        user_qss.write_text(base_qss.read_text())

    return user_qss


class _SingletonMeta(type):
    _instance = None

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        if not cls._instance:
            instance = super().__call__(*args, **kwargs)
            cls._instance = instance
        return cls._instance


@dataclass
class Settings(metaclass=_SingletonMeta):
    max_workers: int = 20
    enable_debug: bool = False
    wrap_lines: bool = True
    view: Dict[str, Any] = field(default_factory=dict[str, Any], repr=False)


@contextmanager
def open_settings(
    mode: Literal['r', 'w'] = 'r',
    *, json_file_path: Optional[Path] = None
) -> Generator[Settings, Any, None]:
    """Open settings context manager.

    ```
    with open_settings(mode='r') as s:
        s.save_to_db = True
        s.focus_tab_on_error = False
    ```
    """
    if json_file_path is None:
        json_file_path = Path(os.environ['TQM_SETTINGS_PATH'])

    try:
        with json_file_path.open() as f:
            settings = Settings(**json.load(f))

    except FileNotFoundError:
        settings = Settings()
        with json_file_path.open('w') as f:
            json.dump(asdict(settings), f, indent=4)

    except Exception as e:
        print('[tqm error]: Invalid settings file. Resetting settings.', e)
        settings = Settings()

    yield settings

    if mode != 'w':
        return

    with json_file_path.open('w') as f:
        json.dump(asdict(settings), f, indent=4)



from __future__ import annotations

from typing import Any, Callable, Optional
from functools import partial


def extract_fn_name(func: Optional[Callable[..., Any]]) -> str:
    """
    Extracts the name of a function for debugging purposes.

    Args:
        func (Optional[Callable[..., Any]]): The function to extract the name from.

    Returns:
        str: The name of the function.

    """
    if not func:
        return ''

    if isinstance(func, partial):
        return func.func.__name__

    try:
        return func.__name__
    except Exception:
        return str(func)



from __future__ import annotations

from typing import Optional

from PySide2.QtWidgets import QFrame, QWidget


class Frame(QFrame):
    """Custom frame for styling purposes."""

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)

        self.setObjectName('TqmFrame')
        self.setFrameStyle(QFrame.StyledPanel)



from __future__ import annotations

from typing import (Any, Set, Dict, List, Generic, TypeVar, Callable, Optional,
                    overload)

from PySide2.QtGui import QColor
from PySide2.QtCore import Qt

from .task import TaskUnit, TaskGroup, TaskExecutable
from ..utils import RandomColor
from ..typings import RGBA, TASK_COLOR
from .task_actions import TaskAction, TaskActionVisibility
from .task_options import ProgressMode, ProgressBarOptions
from .task_callbacks import TaskCallbacks
from .task_predicate import TaskPredicate

TaskType = TypeVar('TaskType', TaskExecutable, TaskGroup)
Builder = TypeVar(
    'Builder',
    bound='_TaskBuilderBase[TaskBuilder | TaskGroupBuilder]'
)


class _TaskBuilderBase(Generic[Builder, TaskType]):
    """Base class for TaskBuilder and TaskGroupBuilder."""
    Actions = TaskActionVisibility

    def __init__(self, name: str):
        self.name = name
        self.color = QColor(230, 230, 230, 255)
        self.comment = ''

        self.parent: Optional[TaskUnit] = None

        self.predicate: Optional[Callable[..., bool]] = None
        self.retry_interval_ms = 2000
        self.predicate_retries = 10

        self.max_retry_failed_attempts = 0

        self.data: Dict[str, Any] = {}

        self.on_start: Optional[Callable[[TaskType], Any]] = None
        self.on_finish: Optional[Callable[[TaskType], Any]] = None
        self.on_completed: Optional[Callable[[TaskType], Any]] = None
        self.on_failed: Optional[Callable[[TaskType], Any]] = None

        self._random_color = RandomColor()
        self._actions: List[TaskAction[TaskType]] = []

    def with_label(self: Builder, name: str) -> Builder:
        """Set the name of the task.

        >>> with_label('My Task')

        """
        self.name = name
        return self

    def with_comment(self: Builder, comment: str) -> Builder:
        """Add a comment to the task."""
        self.comment = comment
        return self

    def with_wait_for(self: Builder, parent: TaskUnit) -> Builder:
        """Wait for another task/group to finish before starting."""
        self.parent = parent
        return self

    def with_retry_failed(self: Builder, max_retries: int = 0) -> Builder:
        """Automatically retry the task if it fails.

        >>> with_retry_on_failed(max_retries=5)

        TODO: Add delay between retries

        """
        self.max_retry_failed_attempts = max_retries
        return self

    def with_predicate(
        self: Builder,
        predicate: Callable[..., bool],
        *,
        max_retries: int = 2,
        retry_interval_ms: int = 2000
    ) -> Builder:
        """Predicate to be executed before the task is started.

        >>> with_predicate(lambda _: os.path.exists('file.txt'), max_retries=2, retry_interval_ms=2000)

        If the predicate returns False, the task will not be executed. If
        `max_retries` is set, the task will be retried that many times every
        `retry_interval_ms` milliseconds.

        """
        self.predicate = predicate
        self.retry_interval_ms = retry_interval_ms
        self.predicate_retries = max_retries
        return self

    def with_action(
        self: Builder,
        label: str,
        execute: Callable[[TaskType], Any],
        visibility: TaskActionVisibility = TaskActionVisibility.ON_COMPLETED
    ) -> Builder:
        """Add an action to the task.

        >>> with_action(
        ...    'Open File',
        ...    lambda t: QDesktopServices.openUrl(f'file://{task.data["file"]}'),
        ...    visibility=TaskActionVisibility.ON_COMPLETED
        ... )
        >>> with_action(
        ...    'Inspect Task',
        ...    lambda task: print(f'Running: {task.state.inspect()}'),
        ...    visibility=TaskActionVisibility.ALWAYS
        ... )

        An action is a context menu item that is displayed when the task is right-clicked.
        Can be called multiple times to add multiple actions.

        Tip: If you need to open a file, use `with_file` instead.

        """
        self._actions.append(TaskAction[TaskType](label, execute, visibility))
        return self

    def with_file_action(
        self: Builder,
        file: str,
        visibility: TaskActionVisibility = TaskActionVisibility.ON_COMPLETED
    ) -> Builder:
        """
        Adds a file menu action to the task, specifying a file to be used when the action is triggered.

        >>> with_file_action(__file__, visibility=TaskActionVisibility.ALWAYS)

        """
        self._actions.append(TaskAction[TaskType]('%file%', lambda _: file, visibility))
        return self

    def with_on_start(self: Builder, on_start: Callable[[TaskType], Any], *args: Any, **kwargs: Any) -> Builder:
        """Callback to be executed before the task is started.

        >>> with_on_start(lambda task: print(task.name))

        NOTE: The first argument is the task itself.

        """
        self.on_start = on_start
        return self

    def with_on_failed(self: Builder, on_failed: Callable[[TaskType], Any]) -> Builder:
        """Callback to be executed if the task fails.

       >>> with_on_failed(lambda task: print(f'Error in {task.name}'))

        NOTE: The first argument is the task itself.

        """
        self.on_failed = on_failed
        return self

    def with_on_finish(self: Builder, on_finish: Callable[[TaskType], Any]) -> Builder:
        """Callback to be executed after the task is finished.

        >>> with_on_finish(lambda task: print(task.name))

        NOTE: The first argument is the task itself.

        """
        self.on_finish = on_finish
        return self

    def with_on_completed(self: Builder, on_completed: Callable[[TaskType], Any]) -> Builder:
        """Callback to be executed when the task is completed.

        >>> with_on_completed(lambda task: print(task.name))

        NOTE: The first argument is the task itself.

        """
        self.on_completed = on_completed
        return self

    def with_data(self: Builder, **kwargs: Any) -> Builder:
        """Set additional data to the task.

        This can be used to store any additional data that is needed for the task.

        >>> with_data(key1='value1', key2='value2')

        NOTE: Make sure the data is JSON serializable as it will be stored as a string.

        """
        self.data.update(kwargs)
        return self

    @overload
    def with_color(self: Builder, color: RGBA) -> Builder:
        """Set the color of the task with an RGBA tuple.

        >>> with_color((255, 0, 0, 70))

        """

    @overload
    def with_color(self: Builder, color: Qt.GlobalColor) -> Builder:
        """Set the color of the task with a Qt.GlobalColor.

        >>> with_color(Qt.red)

        """

    @overload
    def with_color(self: Builder, color: str) -> Builder:
        """Set the color of the task with a string. It can be a color name or a hex value.

        >>> with_color('red')
        >>> with_color('#ff0000')

        """

    @overload
    def with_color(self: Builder) -> Builder:
        """Set the color of the task with a random color."""

    @overload
    def with_color(self: Builder, color: Optional[TASK_COLOR] = None) -> Builder: ...

    def with_color(self: Builder, color: Optional[TASK_COLOR] = None) -> Builder:
        """Set the color of the task.

        >>> with_color('red')
        >>> with_color('#ff0000')
        >>> with_color(Qt.red)
        >>> with_color('rgb(255, 0, 0)')
        >>> with_color((255, 0, 0, 70))
        >>> with_color() # Random color

        """
        if isinstance(color, QColor):
            self.color = color

        elif isinstance(color, (str, Qt.GlobalColor)):
            user_color = QColor(color)
            self.color = user_color if user_color.isValid() else self._random_color.generate()

        elif isinstance(color, tuple):
            self.color = QColor(*color)

        else:
            self.color = self._random_color.generate()

        return self


class TaskBuilder(_TaskBuilderBase['TaskBuilder', TaskExecutable]):
    """
    A builder class for creating Task instances.

    >>> task = TaskBuilder('My Task').build()

    """

    def __init__(self, name: str = ''):
        super().__init__(name)

        self.event: Callable[[TaskExecutable], Any] = lambda task: None

        # Progress bar options
        self.minimum = 0
        self.maximum = 100
        self.show_progress: Optional[bool] = None

    def with_event(
        self,
        event: Callable[[TaskExecutable], Any],
        *,
        show_progress: bool = False
    ) -> TaskBuilder:
        """Set the event of the task.

        >>> with_event(lambda task: print(task.name))
        >>> with_event(show_progress_event, show_progress=True)

        NOTE:
            - The first argument is the task itself.
            - The progress is emitted by the `task.update_progress` signal

        """
        self.event = event
        self.show_progress = show_progress
        return self

    def with_min_max(self, minimum: int = 0, maximum: int = 100) -> TaskBuilder:
        """Minimum and maximum values for the progress bar."""
        self.minimum = minimum
        self.maximum = maximum
        return self

    def build(self) -> TaskExecutable:
        return TaskExecutable(
            name=self.name,
            execute=self.event,
            actions=tuple(self._actions),
            comment=self.comment,
            parent=self.parent,
            color=self.color,
            data=self.data,
            progress_bar=ProgressBarOptions(
                minimum=self.minimum,
                maximum=self.maximum,
                mode=(
                    ProgressMode.DETERMINATE
                    if self.show_progress
                    else ProgressMode.INDETERMINATE
                )
            ),
            predicate=TaskPredicate(
                condition=self.predicate,
                max_retries=self.predicate_retries,
                retry_interval=self.retry_interval_ms,
            ),
            retry_attempts=self.max_retry_failed_attempts,
            callbacks=TaskCallbacks(
                on_start=self.on_start,
                on_finish=self.on_finish,
                on_failed=self.on_failed,
                on_completed=self.on_completed,
            ),
        )


class TaskGroupBuilder(_TaskBuilderBase['TaskGroupBuilder', TaskGroup]):
    """
    A builder class for creating TqmTaskGroup instances.

    >>> group = TaskGroupBuilder('My Group').build()

    """

    def __init__(self, name: str = ''):
        super().__init__(name)

        self.tasks: Set[TaskExecutable] = set()

    def with_tasks(self, *tasks: TaskExecutable) -> TaskGroupBuilder:
        for task in tasks:
            self.tasks.add(task)
        return self

    def build(self) -> TaskGroup:
        group = TaskGroup(
            name=self.name,
            parent=self.parent,
            comment=self.comment,
            color=self.color,
            actions=tuple(self._actions),
            progress_bar=ProgressBarOptions(
                maximum=len(self.tasks),
                mode=ProgressMode.DETERMINATE
            ),
            predicate=TaskPredicate(
                condition=self.predicate,
                max_retries=self.predicate_retries,
                retry_interval=self.retry_interval_ms,
            ),
            callbacks=TaskCallbacks(
                on_start=self.on_start,
                on_finish=self.on_finish,
                on_completed=self.on_completed,
                on_failed=self.on_failed
            )
        )

        group.add_tasks(*self.tasks)
        return group



from __future__ import annotations

import uuid
from typing import (TYPE_CHECKING, Any, Set, Dict, Tuple, Generic, TypeVar,
                    Iterator, Optional, Generator)
from itertools import count
from dataclasses import field, dataclass

from PySide2.QtGui import QColor

from .logger import LOGGER, USER_LEVEL
from .task_state import TaskState
from .task_runner import BaseRunner
from .task_actions import TaskAction
from .task_options import ProgressBarOptions
from .._ui.task_item import TaskItem
from .task_callbacks import TaskCallbacks
from .task_predicate import TaskPredicate

if TYPE_CHECKING:
    from .task import TaskUnit

T = TypeVar('T')
W = TypeVar('W', bound=BaseRunner)

_INDEX_CACHE: Dict[str, Iterator[int]] = {}


def index_gen(suffix: str) -> int:
    if suffix not in _INDEX_CACHE:
        _INDEX_CACHE[suffix] = count(1)
    return next(_INDEX_CACHE[suffix])


@dataclass
class TaskBase(Generic[T, W]):
    """Base class for both Task and TaskGroup entities."""

    # attributes
    name: str = ''
    comment: str = ''
    retry_attempts: int = -1
    exception: Optional[Exception] = field(init=False, default=None)

    # user data container
    data: Dict[str, Any] = field(repr=False, default_factory=dict[str, Any])

    # states
    state: TaskState = field(init=False, default_factory=TaskState, repr=False)

    # events
    actions: Tuple[TaskAction[T], ...] = field(default_factory=tuple, repr=False)
    predicate: TaskPredicate = field(default_factory=TaskPredicate, repr=False)
    callbacks: TaskCallbacks[T] = field(default_factory=TaskCallbacks['T'], repr=False)

    runner: W = field(init=False)

    # UI
    color: QColor = QColor(220, 220, 220, 255)
    item: Optional[TaskItem] = field(init=False, repr=False, default=None)
    progress_bar: ProgressBarOptions = field(default_factory=ProgressBarOptions, repr=False)

    # relationships
    parent: Optional[TaskUnit] = None
    children: Set[TaskUnit] = field(init=False, default_factory=set['TqmTaskUnit'])

    # id
    id: uuid.UUID = field(init=False, default_factory=uuid.uuid4)
    index: int = field(init=False)

    def __post_init__(self, suffix: str = '') -> None:
        """Initialize the task state."""
        self.state.set_inactive()

        self.index = index_gen(suffix)

        if not self.name:
            self.name = f'{suffix}-{str(self.index).zfill(5)}'

        if self.parent:
            LOGGER.debug('Adding %s to parent %s', self.name, self.parent.name)
            self.parent.children.add(self)

        self._initial_attempts = self.retry_attempts

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, TaskBase):
            raise NotImplementedError(f'Cannot compare a task object to {type(value)}')
        return self.id == value.id

    def __lt__(self, other: TaskBase[T, W]) -> bool:
        """Comparison method for the heap queue."""
        return self.index < other.index

    def __str__(self) -> str:
        return f'{self.name}.{self.state} <{self.id}>'

    def log(self, text: str) -> None:
        """Log a message with the task name as prefix."""
        LOGGER.log(USER_LEVEL, f'{self.name}: {text}')

    def get_children(self) -> Set[TaskUnit]:
        """Return all the children of the task recursively."""
        def recurse(task: TaskUnit) -> Generator[TaskUnit, Any, None]:
            for child in filter(lambda t: not t.state.is_inactive, task.children):
                yield child
                yield from recurse(child)
        return set(recurse(self))

    def can_retry(self) -> bool:
        """Check if the task can be retried."""
        return self.retry_attempts > 0

    def retry(self) -> None:
        if self.retry_attempts <= 0:
            raise ValueError('Not enough retry attempts')
        self.retry_attempts -= 1

    def delete(self, comment: str = '') -> None:
        """Mark the task as deleted and clean up resources."""
        self.predicate.delete()
        self.callbacks.delete()
        self.state.set_deleted(comment)

        if self.parent:
            self.parent.children.remove(self)
            self.parent = None

    def reset(self, comment: str = '', reset_attempts: bool = False) -> None:
        self.state.set_inactive(comment)

        if reset_attempts:
            self.retry_attempts = self._initial_attempts

        self.predicate.reset()
        self.exception = None

    def set_failed(self, exception: Optional[Exception] = None, comment: str = '') -> None:
        # XXX should we override previous exception with None?
        if exception is not None:
            self.exception = exception
        self.state.set_failed(comment)

    def inspect(self) -> Dict[str, Any]:
        return {
            'name': self.name,
            'id': str(self.id),
            'retry_attempts': self.retry_attempts,
            'parent': str(self.parent or ''),
            'children': [str(child) for child in self.children],
            'data': self.data,
            'comment': self.comment,
            'callbacks': self.callbacks.inspect(),
            'runner': str(self.runner),
            'state': self.state.inspect(),
            'actions': [action.inspect() for action in self.actions],
            'progress_bar_options': self.progress_bar.inspect(),
            'predicate': self.predicate.inspect(),
            'exception': str(self.exception or ''),
        }



from __future__ import annotations

import os
import logging
from typing import Dict, Optional

from PySide2.QtCore import Qt, QDir, Slot, QPoint, QDateTime
from PySide2.QtWidgets import (QAction, QWidget, QFileDialog, QVBoxLayout,
                               QPlainTextEdit)

from ..utils import RandomColor
from .font_loader import get_monospace_font
from .._core.logger import LOGGER
from .._core.settings import open_settings

LOG_COLORS = {
    'DEBUG': 'aqua',
    'INFO': 'white',
    'WARNING': 'orange',
    'ERROR': 'red',
    'CRITICAL': 'magenta',
}


class TasksLog(QWidget):
    def __init__(self, should_wrap: bool = False, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self._logs = QPlainTextEdit()
        self._logs.setReadOnly(True)
        self._logs.setObjectName('TqmTasksLogs')
        self._logs.setFont(get_monospace_font(14))
        self._logs.setContextMenuPolicy(Qt.CustomContextMenu)
        self._logs.customContextMenuRequested.connect(self._on_contextMenuEvent)

        self._set_line_wrapping(should_wrap)

        self._enable_debug = QAction('Enable Debug', self)
        self._enable_debug.setCheckable(True)
        self._enable_debug.toggled.connect(self._on_enable_debug)

        self._wrap_lines = QAction('Wrap Lines', self)
        self._wrap_lines.setCheckable(True)
        self._wrap_lines.toggled.connect(self._on_wrap_lines)

        self._clear_console = QAction('Clear', self)
        self._clear_console.triggered.connect(self.on_clear_logs)

        self._save_console = QAction('Save', self)
        self._save_console.triggered.connect(self.on_save_logs)

        layout = QVBoxLayout()
        layout.addWidget(self._logs)
        self.setLayout(layout)

        self._random_color = RandomColor()
        self._thread_colors: Dict[str, str] = {}

    def _on_contextMenuEvent(self, point: QPoint) -> None:

        menu = self._logs.createStandardContextMenu()
        menu.addSection('Console Options')
        menu.addAction(self._wrap_lines)
        menu.addAction(self._enable_debug)
        menu.addAction(self._clear_console)
        menu.addAction(self._save_console)
        menu.exec_(self._logs.mapToGlobal(point))

    def _set_line_wrapping(self, state: bool) -> None:
        self._logs.setLineWrapMode(
            QPlainTextEdit.WidgetWidth if state else QPlainTextEdit.NoWrap
        )

    @Slot()
    def _on_wrap_lines(self, state: bool) -> None:
        self._set_line_wrapping(state)
        with open_settings(mode='w') as s:
            s.wrap_lines = state

    @Slot(int)
    def _on_enable_debug(self, state: bool) -> None:
        LOGGER.widget.setLevel(logging.DEBUG if state else logging.INFO)
        with open_settings(mode='w') as s:
            s.enable_debug = state

    @Slot()
    def on_save_logs(self) -> None:
        now = QDateTime.currentDateTime().toString('yyyy-MM-dd_hh-mm-ss')
        report_file = f'{QDir.homePath()}/{now}_tqm_logs.csv'

        filename, _ = QFileDialog.getSaveFileName(
            self, 'Save log', report_file, 'Text files (*.txt);;All files (*)'
        )

        if not filename:
            return

        with open(filename, 'w') as f:
            f.write(self._logs.toPlainText())

    @Slot()
    def on_clear_logs(self) -> None:
        self._logs.clear()

    def log(self, text: str, level_name: str = 'INFO', thread: str = '') -> None:

        if level_name == 'USER':
            if thread in self._thread_colors:
                color = self._thread_colors[thread]
            else:
                color = self._random_color.generate().name()
                self._thread_colors[thread] = color

        else:
            color = LOG_COLORS.get(level_name, 'white')

        text = text.replace(' ', '&nbsp;')
        self._logs.appendHtml(f'<font color="{color}">{text}</font>')
        self._logs.verticalScrollBar().setValue(
            self._logs.verticalScrollBar().maximum()
        )



from __future__ import annotations

import os
import contextlib
from typing import Any, Set, Dict, List, Callable, Optional
from functools import partial

from PySide2.QtCore import Qt, Slot, QTimer, Signal, QObject

from .task import TaskUnit, TaskExecutable
from .queue import TasksQueue, TaskNotFoundError
from .logger import LOGGER
from .shutdown import ShutdownThread
from ..exceptions import (TaskError, TaskParentError, TaskAlreadyInQueue,
                          TaskPredicateError)
from .task_runner import RunnerSignals
from .task_predicate import PredicateEventType
from .threadpool_interface import ThreadPoolWrapper, ThreadPoolInterface


class _ExecutorCallbacks(RunnerSignals):

    # ui
    task_added = Signal(object)
    task_removed = Signal(object)

    status_updated = Signal(dict)
    system_idle = Signal()

    def __init__(self, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)

    def on_task_added(self, callback: Callable[[TaskUnit], Any]) -> None:
        """On task added callback.

        ```
        on_task_added(lambda t: print(f'Task added: {t.name}'))
        ```

        """
        self.task_added.connect(callback, Qt.QueuedConnection)

    def on_task_started(self, callback: Callable[[TaskUnit], Any]) -> None:
        """On task started callback.


        ```
        on_task_started(lambda t: print(f'Task started: {t.name}'))
        ```

        """
        self.runner_started.connect(callback, Qt.QueuedConnection)

    def on_task_completed(self, callback: Callable[[TaskUnit], Any]) -> None:
        """On task completed callback.

        ```
        on_task_completed(lambda t: print(f'Task completed: {t.name}'))
        ```

        """
        self.runner_completed.connect(callback, Qt.QueuedConnection)

    def on_task_finished(self, callback: Callable[[TaskUnit], Any]) -> None:
        """On task finished callback.

        ```
        on_task_finished(lambda t: print(f'Task finished: {t.name}'))
        ```

        """
        self.runner_finished.connect(callback, Qt.QueuedConnection)

    def on_task_failed(self, callback: Callable[[TaskUnit], Any]) -> None:
        """On task failed callback.

        ```
        on_task_failed(lambda t: print(f'Task failed: {t.name}'))
        ```

        """
        self.runner_failed.connect(callback, Qt.QueuedConnection)

    def on_task_removed(self, callback: Callable[[TaskUnit], Any]) -> None:
        """On task removed callback.

        ```
        on_task_removed(lambda t: print(f'Task removed: {t.name}'))
        ```

        """
        self.task_removed.connect(callback, Qt.QueuedConnection)

    def on_system_idle(self, callback: Callable[..., Any]) -> None:
        """On workers idle callback.

        NOTE: This signal is primarily intended for testing and debugging.
        For production code, consider using task groups with completion signals.

        The signal is emitted when the system has been idle (no running, waiting,
        or queued tasks) for a short period (100ms by default).

        ```
        on_system_idle(lambda: print('System currently idle'))
        ```
        """
        self.system_idle.connect(callback, Qt.QueuedConnection)

    def on_status_updated(self, callback: Callable[[Dict[str, Any]], Any]) -> None:
        """On status updated callback.

        ```
        on_status_updated(lambda s: print(s))
        ```

        """
        self.status_updated.connect(callback, Qt.QueuedConnection)


class _ExecutorBlocker(QObject):
    predicate_successful = Signal(object)
    predicate_failed = Signal(object)

    def __init__(self, executor: TaskExecutor, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)
        self._executor = executor

    def _handle_predicate(
        self,
        task: TaskUnit,
        predicate_event: PredicateEventType
    ) -> None:

        if predicate_event == PredicateEventType.SUCCESS:
            LOGGER.info(f"{task.name} predicate passed")
            self.predicate_successful.emit(task)

        elif predicate_event == PredicateEventType.FAIL:
            task.exception = TaskPredicateError(task.name)
            self.predicate_failed.emit(task)

        elif predicate_event == PredicateEventType.RETRY:
            task.state.set_retrying(f'Attempts left: {task.predicate.retry_left}')
            LOGGER.info(
                f'{task.name} retrying in {task.predicate.retry_interval}ms. '
                f'Retries left: {task.predicate.retry_left}'
            )
        else:
            raise TaskPredicateError('Unknown predicate event')

    def _block_task_with_predicate(self, task: TaskUnit) -> bool:
        """Blocks the given task if its predicate is not resolved."""
        if not task.predicate.condition or task.predicate.condition():
            return False

        LOGGER.info(f'{task.name} predicate failed')
        task.state.set_blocked('Predicate failed')
        return True

    def _block_task_with_parent(self, task: TaskUnit) -> bool:
        """Blocks the given task if its parent task is not completed."""
        if not task.parent:
            return False

        if not task.parent.state.is_completed and not task.state.is_inactive:
            LOGGER.info(f'{task.name} waiting for {task.parent.name}')
            task.state.set_blocked(f'Waiting for parent: {task.parent.name}')
            return True

        return False

    def should_block(self, task: TaskUnit) -> bool:
        if self._block_task_with_predicate(task):
            self._executor.queue.suspend(task)
            task.predicate.evaluate(partial(self._handle_predicate, task))
            return True

        if self._block_task_with_parent(task):
            self._executor.queue.suspend(task)
            return True

        return False


class _ExecutorStateTracker(QObject):
    def __init__(self, executor: TaskExecutor, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)
        self._executor = executor
        self._state_counts = {
            'waiting': 0,
            'running': 0,
            'completed': 0,
            'failed': 0,
        }

        self._idle_timer: Optional[QTimer] = None
        self._idle_timeout = int(os.getenv('TQM_IDLE_TIMEOUT', 1000))

    @property
    def running_tasks(self) -> int:
        return self._state_counts['running']

    def update_status(self, old_state: str, new_state: str) -> None:
        if old_state in self._state_counts:
            self._state_counts[old_state] -= 1

        if new_state in self._state_counts:
            self._state_counts[new_state] += 1

        status: Dict[str, int] = {
            'Running': self._state_counts['running'],
            'Completed': self._state_counts['completed'],
            'Failed': self._state_counts['failed'],
        }

        self._executor.callbacks.status_updated.emit(status)
        self._handle_idle()

    def is_idle(self) -> bool:
        return (
            self._state_counts['running'] == 0 and
            self._state_counts['waiting'] == 0 and
            self._executor.queue.size() == 0 and
            self._executor.queue.size_deferred() == 0
        )

    @Slot(dict)
    def _emit_idle(self) -> None:
        self._executor.callbacks.system_idle.emit()

    def _handle_idle(self) -> None:
        if self.is_idle():

            if not self._idle_timer or not self._idle_timer.isActive():
                self._idle_timer = QTimer()
                self._idle_timer.setSingleShot(True)
                self._idle_timer.timeout.connect(self._emit_idle)
                self._idle_timer.start(self._idle_timeout)

        elif self._idle_timer and self._idle_timer.isActive():
            self._idle_timer.stop()


class TaskExecutor(QObject):

    def __init__(
        self,
        max_workers: int = 20,
        threadpool: Optional[ThreadPoolInterface] = None,
        parent: Optional[QObject] = None
    ):
        super().__init__(parent)
        LOGGER.info('Initializing TasksManager')

        self.queue = TasksQueue()
        self.callbacks = _ExecutorCallbacks()
        self.status_tracker = _ExecutorStateTracker(self)

        self._max_workers = max_workers

        self._threadpool = threadpool or ThreadPoolWrapper(self)
        self._threadpool.setMaxThreadCount(self.max_workers+1)
        self._shutdown_thread = ShutdownThread(self._threadpool)

        self._blocker = _ExecutorBlocker(self)
        self._blocker.predicate_failed.connect(self._on_task_failed)
        self._blocker.predicate_successful.connect(self._predicate_success)

        self.registry: Set[TaskUnit] = set()

        self._is_shutting_down = False

    @property
    def max_workers(self) -> int:
        return self._max_workers

    @max_workers.setter
    def max_workers(self, value: int) -> None:
        self._max_workers = max(2, value)

    def set_max_workers(self, value: int) -> None:
        self.max_workers = value
        self._threadpool.setMaxThreadCount(self.max_workers+1)

    def _predicate_success(self, task: TaskUnit) -> None:
        self.queue.promote_to_main(task)
        self.start_workers()

    def _remove_and_cleanup_task(self, task: TaskUnit) -> None:
        task.delete()
        for child in task.children.copy():
            self.remove_task(child)
        self.callbacks.task_removed.emit(task)

    def remove_task(self, task: TaskUnit) -> None:
        """Attempt to remove a task from the queue.

        Tasks that depend on the removed task will be also removed since they will
        never be executed.

        Raises:
            TaskNotFound: If the task is not found in the queue.

        """
        if task.state.is_inactive:
            self._remove_and_cleanup_task(task)
            return

        if task not in self.registry:
            return

        if not task.state.is_removable:
            raise TaskError(
                f'Cannot remove task: {task.name}. '
                f'State "{task.state.current}" is not removable'
            )

        with contextlib.suppress(TaskNotFoundError):
            self.queue.remove_task(task)

        self.registry.remove(task)

        self._remove_and_cleanup_task(task)

    def retry_task(self, task: TaskUnit) -> None:
        """Retry a failed task.

        This method retries a failed task by re-adding it to the queue.
        The task must be in the deferred queue and have retries left.

        Raises:
            DeferredTaskNotFound: If the task is not found in the deferred queue.

        """
        task.retry_attempts -= 1

        task.reset('Reset & Retry')

        for child in task.children.copy():
            if child.state.is_failed:
                self.retry_task(child)

        if (
            isinstance(task, TaskExecutable)
            and (task.group and task.group.state.is_failed)
        ):
            self.retry_task(task.group)

        self._initialize_task(task)
        self._start_worker()

    @Slot(object)
    def _on_task_failed(self, task: TaskUnit, exception: Optional[Exception] = None) -> None:
        """Handle a failed task.

        If the task can be retried, it will be re-added to the queue.

        Returns:
            str: The exception message if any

        """
        if task.can_retry():
            self.retry_task(task)
            return

        LOGGER.error(f'{task.name} failed: {exception}')

        for child in filter(lambda t: t.state.is_blocked, task.children):
            self._on_task_failed(child, TaskParentError(f'Parent {task.name} failed.'))

        if self.queue.is_task_deferred(task):
            self.queue.remove_from_deferred(task)

        task.set_failed(exception, str(exception))

    @Slot(object)
    def _on_task_completed(self, task: TaskUnit, *, autostart: bool = True) -> None:
        task.state.set_completed()
        LOGGER.info(f'{task.name} completed')

        self.callbacks.runner_completed.emit(task)

        for child in filter(lambda t: t.state.is_blocked, task.children):
            child.state.set_waiting('Unblock from parent')
            self.queue.promote_to_main(child)

        if autostart:
            self.start_workers()

    @Slot(object)
    def _on_task_started(self, task: TaskUnit) -> None:
        self.callbacks.runner_started.emit(task)
        task.state.set_running()

    @Slot(list)
    def _on_add_tasks(self, tasks: List[TaskUnit]) -> None:
        for task in tasks:
            if task not in self.registry:
                self.add_task(task)

        self.start_workers()

    def _initialize_task(self, task: TaskUnit) -> None:

        self.queue.enqueue(task)

        task.state.set_waiting()

        signals = task.runner.signals
        signals.runner_failed.connect(self._on_task_failed, Qt.QueuedConnection)
        signals.runner_completed.connect(self._on_task_completed, Qt.QueuedConnection)
        signals.runner_started.connect(self._on_task_started, Qt.QueuedConnection)
        signals.runner_finished.connect(self.callbacks.runner_finished.emit, Qt.QueuedConnection)
        signals.group_task_added.connect(self._on_add_tasks)

    def add_task(self, task: TaskUnit) -> TaskUnit:
        if self._is_shutting_down:
            LOGGER.warning('Task manager is shutting down. Cannot add new task')
            return task

        if task in self.registry:
            raise TaskAlreadyInQueue(f'Task "{task}" already in queue')

        if isinstance(task, TaskExecutable):
            task.state.register_state_change_callback(self.status_tracker.update_status)

        LOGGER.info('Adding task to queue: %s', task.name)

        self.registry.add(task)
        self._initialize_task(task)

        return task

    def _start_worker(self) -> None:
        task = self.queue.dequeue()

        if self._blocker.should_block(task):
            LOGGER.info(f'{task.name} is on hold')
            return

        self._threadpool.start(task.runner)

    def start_workers(self) -> None:
        """Start the workers."""
        while self.queue.size() and self._threadpool.activeThreadCount() <= self.max_workers:
            self._start_worker()

    def get_all_tasks(self) -> Set[TaskUnit]:
        return self.registry

    def shutdown(self) -> None:
        """Shutdown the task manager.

        This method clears the queue and waits for workers to finish their tasks.

        Note: Workers will complete their current tasks before shutting down.
        A new thread handles the shutdown.

        """
        if not self.status_tracker.running_tasks:
            return

        LOGGER.info('Removing all tasks from queue and waiting for workers to finish...')

        while self.queue.size():
            self.queue.dequeue()

        self._is_shutting_down = True
        self._shutdown_thread.run()
        self._is_shutting_down = False



"""Shutdown tasks for the task runner.

This module contains the shutdown tasks for the task runner. The shutdown tasks
will wait (blocking the main thread) for all tasks to complete before shutting down
the workers.

"""
from __future__ import annotations

from typing import Optional

from PySide2.QtCore import Signal, QObject, QThread, QEventLoop
from PySide2.QtWidgets import QProgressDialog

from .threadpool_interface import ThreadPoolInterface


class ShutdownWorker(QObject):
    finished = Signal()

    def __init__(self, threadpool: ThreadPoolInterface) -> None:
        super().__init__()
        self.threadPool = threadpool

    def run(self) -> None:
        self.threadPool.waitForDone()
        self.finished.emit()


class ShutdownThread(QObject):
    def __init__(self, threadpool: ThreadPoolInterface) -> None:
        super().__init__()

        self.threadPool = threadpool
        self.progressDialog: Optional[QProgressDialog] = None

    def run(self) -> None:
        self.shutdownThread = QThread()
        self.shutdownWorker = ShutdownWorker(self.threadPool)

        self.shutdownWorker.moveToThread(self.shutdownThread)
        self.shutdownWorker.finished.connect(self.onShutdownComplete)

        self.shutdownThread.started.connect(self.shutdownWorker.run)

        self.progressDialog = QProgressDialog('Shutting down, please wait...', '', 0, 0)
        self.progressDialog.setCancelButton(None)
        self.progressDialog.setModal(True)
        self.progressDialog.show()

        self.shutdownThread.start()
        event_loop = QEventLoop()
        self.shutdownThread.finished.connect(event_loop.quit)
        event_loop.exec_()

    def onShutdownComplete(self) -> None:
        self.shutdownThread.quit()
        self.shutdownThread.wait()
        if self.progressDialog:
            self.progressDialog.accept()



"""TQM Demo Module

This module provides a comprehensive demo of TQM (Task Queue Manager) features.
It can be run as a standalone application to showcase TQM capabilities.

Usage:
    Import the module and run the demo function:
    >>> from tqm.demo import run_demo
    >>> run_demo()

    Or simply run the module as a script from the application environment:
    $ tqm-demo
"""

from __future__ import annotations

import sys
import time
import random
import tempfile
from functools import partial
from itertools import count

from PySide2.QtGui import QDesktopServices
from PySide2.QtCore import QUrl
from PySide2.QtWidgets import QMainWindow, QMessageBox, QApplication

from tqm import (TQManager, TaskBuilder, TaskExecutable, TaskGroupBuilder,
                 TaskActionVisibility)


class DemoTasks:
    """Collection of demo tasks for TQM."""

    def __init__(self, manager: TQManager):
        self.manager = manager

    def add_simple_tasks(self):
        """Add simple tasks that complete quickly."""
        self.manager.add_event(
            execute=lambda t: time.sleep(2.5),
            label='Simple Task',
            comment='A basic task that completes quickly'
        )

    def add_progress_tasks(self):
        """Add tasks with progress reporting."""

        def task_with_progress(task: TaskExecutable, steps: int = 10):
            """Task that reports progress as it executes."""
            for i in range(steps):
                progress = (i + 1) / steps * 100
                task.emit_progress(progress)
                task.log(f"Progress: {int(progress)}%")
                time.sleep(0.5)

        self.manager.add_event(
            execute=partial(task_with_progress, steps=10),
            show_progress=True,
            label='Progress Task (10 steps)',
            comment='Task that reports progress'
        )

        self.manager.add_event(
            execute=partial(task_with_progress, steps=5),
            show_progress=True,
            label='Progress Task (5 steps)',
            comment='Another task with progress'
        )

    def add_task_with_action(self):
        """Add a task that creates a file and provides actions to open it."""

        def create_file_task(task: TaskExecutable):
            content = 'This file was created by TQM Demo.\n\n'
            content += 'The Queue Task Manager (TQM) is a powerful tool for managing tasks.\n'
            content += 'You can use it to run multiple tasks in parallel, track their progress,\n'
            content += 'and manage dependencies between them.'

            with tempfile.NamedTemporaryFile('w', delete=False) as f:
                f.write(content)

            # saved file in internal data
            task.data['file_path'] = f.name
            task.log(f"Created file: {f.name}")
            time.sleep(1)  # Simulate some work

        def open_file_action(task: TaskExecutable):
            """Action to open the file."""
            file_path = task.data.get('file_path', '')
            if file_path:
                QDesktopServices.openUrl(QUrl.fromLocalFile(file_path))

        task = (
            TaskBuilder('Create File Task')
            .with_event(create_file_task)
            .with_action(
                'Open File',
                open_file_action,
                visibility=TaskActionVisibility.ON_COMPLETED
            )
            .with_comment('Creates a text file and provides actions to open it')
            .build()
        )

        self.manager.add_tasks(task)

    def add_failing_task(self):
        """Add a task that fails and can be retried."""

        def failing_task(task: TaskExecutable, should_succeed: bool):
            task.log('This task will intentionally fail')
            time.sleep(3)
            if not should_succeed:
                raise RuntimeError('Task failed intentionally. You can retry it!')

        attempts = count()

        task = (
            TaskBuilder('Failing Task')
            .with_event(lambda t: failing_task(t, next(attempts) == 1))
            .with_comment(
                'This task fails intentionally to demonstrate error handling'
            )
            .build()
        )

        self.manager.add_tasks(task)

    def add_task_group_sequence(self):
        """Add a group of related tasks."""

        def task_executor(task: TaskExecutable, delay: float = 1.0):
            task.log(f"Executing {task.name}")
            time.sleep(delay)

        with self.manager.create_group('Sequence Order Group') as group:

            # Create a chain of dependent tasks
            task1 = (
                TaskBuilder('Prerequisite Task')
                .with_event(partial(task_executor, delay=2.0))
                .with_comment('This task must complete before the next ones can start')
                .build()
            )

            task2 = (
                TaskBuilder('Dependent Task 1')
                .with_event(partial(task_executor, delay=2.8))
                .with_wait_for(task1)
                .with_comment('This task waits for Prerequisite Task to complete')
                .build()
            )

            task3 = (
                TaskBuilder('Dependent Task 2')
                .with_event(partial(task_executor, delay=2.2))
                .with_wait_for(task2)
                .with_comment('This task waits for Dependent Task 1 to complete')
                .build()
            )

            group.add_tasks(task1, task2, task3)

    def add_random_color_tasks(self):
        """Add tasks with random colors."""

        for i in range(3):
            task = (
                TaskBuilder(f"Colored Task {i+1}")
                .with_event(lambda t: time.sleep(random.uniform(0.5, 1.5)))
                .with_color()
                .build()
            )
            self.manager.add_tasks(task)

    def add_task_with_callbacks(self):
        """Add a task with various lifecycle callbacks."""

        def on_start(task: TaskExecutable):
            task.log('Task started callback triggered')

        def on_completed(task: TaskExecutable):
            task.log('Task completed callback triggered')

        def on_finish(task: TaskExecutable):
            task.log('Task finished callback triggered')

        task = (
            TaskBuilder('Callback Demo Task')
            .with_event(lambda t: (t.log('Main task execution'), time.sleep(2.0)))
            .with_on_start(on_start)
            .with_on_completed(on_completed)
            .with_on_finish(on_finish)
            .with_comment('Demonstrates task lifecycle callbacks')
            .build()
        )

        self.manager.add_tasks(task)

    def add_complex_group(self):
        """Add a complex group with nested structures and dependencies."""

        # Create main group
        main_group = (
            TaskGroupBuilder('Complex Process Group')
            .with_comment('A complex group demonstrating advanced TQM features')
            .with_color('lightskyblue')
            .build()
        )

        # First sub-task directly in group
        init_task = (
            TaskBuilder('Initialize Process')
            .with_event(lambda t: (t.log('Initializing process...'), time.sleep(1.8)))
            .with_comment('First step in the complex process')
            .build()
        )

        # Create a nested group for parallel execution
        parallel_group = (
            TaskGroupBuilder('Parallel Tasks')
            .with_comment('These tasks can run in parallel')
            .with_wait_for(init_task)
            .with_color()
            .build()
        )

        # Create tasks for parallel execution
        for i in range(3):
            task = (
                TaskBuilder(f"Parallel Task {i+1}")
                .with_event(lambda t: (t.log(f"Running {t.name}..."), time.sleep(random.uniform(0.5, 2.0))))
                .with_comment(f"Parallel execution task {i+1}")
                .build()
            )
            parallel_group.add_tasks(task)

        # Final task to finish process
        finish_task = (
            TaskBuilder('Finalize Process')
            .with_event(lambda t: (t.log('Finalizing process...'), time.sleep(1.8)))
            .with_comment('Last step in the complex process')
            .with_wait_for(parallel_group)
            .build()
        )

        # # Add all tasks to the main group
        main_group.add_tasks(init_task, finish_task)

        # Add groups to manager
        self.manager.add_tasks(main_group, parallel_group)


def setup_manager() -> TQManager:
    """
    Run the TQM demo application.

    Returns:
        The QApplication instance if standalone=True, otherwise None.
    """

    # Setup task manager
    task_manager = TQManager(app_name='tqm_demo')
    task_manager.setWindowTitle('TQM Demo - Queue Task Manager')

    # Set up demo tasks
    demo_tasks = DemoTasks(task_manager)

    # Add various types of tasks
    demo_tasks.add_simple_tasks()
    demo_tasks.add_progress_tasks()
    demo_tasks.add_task_with_action()
    demo_tasks.add_failing_task()
    demo_tasks.add_task_group_sequence()
    demo_tasks.add_random_color_tasks()
    demo_tasks.add_task_with_callbacks()
    demo_tasks.add_complex_group()

    QMessageBox.information(
        task_manager,
        'TQM Demo',
        'Welcome to the TQM Demo!\n\n'
        'This demo showcases various features of the Queue Task Manager:\n'
        '• Simple tasks\n'
        '• Tasks with progress reporting\n'
        '• File-related tasks with context menu actions\n'
        '• Task failure and retry\n'
        '• Task groups\n'
        '• Task dependencies\n'
        '• Task colors\n'
        '• Complex nested groups\n'
        '• Task lifecycle callbacks\n\n'
        "Click 'Start Workers' to begin executing the tasks.\n"
        'You can interact with tasks by right-clicking on them.'
    )

    # Display the task manager
    return task_manager


def run_demo():
    """
    Main entry point when run as a script.
    """
    app = QApplication(sys.argv)
    app.setStyle('Fusion')

    window = QMainWindow()
    window.setCentralWidget(setup_manager())
    window.show()

    sys.exit(app.exec_())
